from ._core_base import (_BaseObject, _BaseProperty, ksplit, BaseData, kjoin,  open_object, ChildrenCapabilityConfig,
                            ChildrenCapability)

from ._core_device import BaseDevice 
from ._core_node import BaseNode
from ._core_rpc import BaseRpc  
from ._core_interface import BaseInterface  
from ._class_recorder import KINDS, get_class, record_class

from ._core_pydantic import _default_walk_set

from ._core_obj_dict import ObjDict
from pydantic import create_model 


                      
class ManagerConfig(_BaseObject.Config, ChildrenCapabilityConfig):
    kind: KINDS = KINDS.MANAGER.value
    type: str = "Base"

class ManagerProperty(_BaseProperty):    
    fbuild = None    
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self
    
    def __call__(self, func):
        """ The call is used has fget decorator """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, device):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, device)

def open_manager(cfgfile, path=None, prefix="", key=None, default_type=None, **kwargs):
    """ Open a manager from a configuration file 

        
        Args:
            cfgfile: relative path to one of the $CFGPATH or absolute path to the yaml config file 
            key: Key of the created Manager 
            path (str, int, optional): 'a.b.c' will loock to cfg['a']['b']['c'] in the file. If int it will loock to the Nth
                                        element in the file
            prefix (str, optional): additional prefix added to the name or key

        Output:
            manager (BaseManager subclass) :tanciated Manager class     
    """
    kwargs.setdefault("kind", KINDS.MANAGER)
    return open_object(
                cfgfile, 
                path=path, prefix=prefix, 
                key=key, default_type=default_type, 
                **kwargs
            ) 



class ManagerDict(ObjDict):
   class Config(ObjDict.Config):
        kinds = set([KINDS.MANAGER])
        default_kind = KINDS.MANAGER

@record_class        
class BaseManager(_BaseObject, ChildrenCapability):
    Property  = ManagerProperty
    Config = ManagerConfig
    Data = BaseData
    Device = BaseDevice
    Interface = BaseInterface
    Node = BaseNode
    Rpc = BaseRpc
    Dict = ManagerDict
    
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._localdata is None:
            self._localdata = {}
    

    @property
    def devices(self):
        return self.find( BaseDevice )
        # for obj in self.__dict__.values():
        #     if isinstance(obj, BaseDevice):
        #         yield obj
    
               
   

    def connect(self) -> None:
        """ Connect all devices """
        for device in self.devices:
            device.connect()
    
    def disconnect(self) -> None:
        """ disconnect all devices """
        for device in self.devices:
            device.disconnect()                
                
                    
    @classmethod
    def parse_config(cls, config, **kwargs):
        if isinstance(config, dict):
            kwargs = {**config, **kwargs}
            config = None
           
        return super().parse_config(config, **kwargs)
        
    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()
    
        



import math
import ast
import operator as op


_operators2method = {
    ast.Add: op.add, 
    ast.Sub: op.sub, 
    ast.BitXor: op.xor, 
    ast.Or:  op.or_, 
    ast.And: op.and_, 
    ast.Mod:  op.mod,
    ast.Mult: op.mul,
    ast.Div:  op.truediv,
    ast.Pow:  op.pow,
    ast.FloorDiv: op.floordiv,              
    ast.USub: op.neg, 
    ast.UAdd: lambda a:a  
}

class DataEval:
    """ Basic parser with local variable and math functions 
    
    Args:
       vars (mapping): mapping object where obj[name] -> numerical value 
       math (bool, optional): if True (default) all math function are added in the same name space
       
    Example:
       
       data = {'r': 3.4, 'theta': 3.141592653589793}
       parser = DataEval(data)
       assert parser.parse('r*cos(theta)') == -3.4
       data['theta'] =0.0
       assert parser.parse('r*cos(theta)') == 3.4
    """
        
    
    
    def __init__(self, vars: dict, math=True):
        self._vars = vars
        if not math:
            self._alt_name = self._no_alt_name
        
            
    def _Name(self, name):
        try:
            return  self._vars[name]
        except KeyError:
            return self._alt_name(name)
                
    @staticmethod
    def _alt_name(name):
        if name.startswith("_"):
            raise NameError(f"{name!r}") 
        try:
            return  getattr(math, name)
        except AttributeError:
            raise NameError(f"{name!r}") 
    
    @staticmethod
    def _no_alt_name(name):
        raise NameError(f"{name!r}") 
    
    @staticmethod
    def Call(f,args):
        return f(*args)
    
    def eval_(self, node):
        if isinstance(node, ast.Expression):
            return self.eval_(node.body)
        if isinstance(node, ast.Num): # <number>
            return node.n
        if isinstance(node, ast.Name):
            return self._Name(node.id) 
        if isinstance(node, ast.BinOp):            
            method = _operators2method[type(node.op)]                      
            return method( self.eval_(node.left), self.eval_(node.right) )            
        if isinstance(node, ast.UnaryOp):             
            method = _operators2method[type(node.op)]  
            return method( self.eval_(node.operand) )
        if isinstance(node, ast.Attribute):
            return getattr(self.eval_(node.value), node.attr)
            
        if isinstance(node, ast.Call):            
            return self.eval_(node.func)( 
                      *(self.eval_(a) for a in node.args),
                      **{k.arg:self.eval_(k.value) for k in node.keywords}
                     )
        else:
            raise TypeError(node)
    
    def eval(self, expr):
        return  self.eval_(ast.parse(expr, mode='eval'))          
    



class ExpEval:
    """ Basic parser from a set of data
    
    Args:
       exp (str): expression to eval 
       math (bool, optional): if True (default) all math function are added in the same name space
       
    Example:
       
       
    """
    
    def __init__(self, exp: str, math=True):        
        self._exp = ast.parse(exp, mode='eval')                 
        if not math:
            self._alt_name = self._no_alt_name
        
            
    def _Name(self, name, vars):
        try:
            return vars[name]
        except KeyError:
            return self._alt_name(name)
                
    @staticmethod
    def _alt_name(name):
        if name.startswith("_"):
            raise NameError(f"{name!r}") 
        try:
            return  getattr(math, name)
        except AttributeError:
            raise NameError(f"{name!r}") 
    
    @staticmethod
    def _no_alt_name(name):
        raise NameError(f"{name!r}") 
    
    @staticmethod
    def Call(f,args):
        return f(*args)
    
    def eval_(self, node, vars):
        if isinstance(node, ast.Expression):
            return self.eval_(node.body, vars)
        if isinstance(node, ast.Num): # <number>
            return node.n
        if isinstance(node, ast.Name):
            return self._Name(node.id, vars) 
        if isinstance(node, ast.BinOp):            
            method = _operators2method[type(node.op)]                      
            return method( self.eval_(node.left, vars), self.eval_(node.right, vars) )            
        if isinstance(node, ast.UnaryOp):             
            method = _operators2method[type(node.op)]  
            return method( self.eval_(node.operand, vars) )
        if isinstance(node, ast.Attribute):
            return getattr(self.eval_(node.value, vars), node.attr)
            
        if isinstance(node, ast.Call):            
            return self.eval_(node.func, vars)( 
                      *(self.eval_(a, vars) for a in node.args),
                      **{k.arg:self.eval_(k.value, vars) for k in node.keywords}
                     )                       
        else:
            raise TypeError(node)
    
    def eval(self, vars):
        return  self.eval_(self._exp, vars)      




if __name__=="__main__":    
    assert DataEval({"x":4.5}).eval('x*2') == 9
    assert DataEval({"x":4}).eval('cos(pi)') == -1.0
        
    data = {'r': 3.4, 'theta': 3.141592653589793}
    parser = DataEval(data)
    assert parser.eval('r*cos(theta)') == -3.4
    data['theta'] = 0.0
    assert parser.eval('r*cos(theta)') == 3.4
    assert DataEval(globals()).eval('math.pi') == math.pi
    
    assert DataEval({'f':lambda x,n=10: x*n}).eval('f(2,20)') == 40
    assert DataEval({'f':lambda x,n=10: x*n}).eval('f(2,n=20)') == 40  
    
    
    assert ExpEval('4*x').eval({'x':5}) == 20
      
    print("OK")


from PyQt5 import  uic
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QComboBox, QWidget
from .base import get_widget_types, BaseUiLinker, get_widget_factory, DEFAULT, LayoutLinker
from pydantic import BaseModel
from typing import Optional, List
from pydevmgr_core import BaseDevice
from .io import find_ui


class SwitchUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('switch.ui'), self)
        
class SwitchLinker(BaseUiLinker):
    Widget = SwitchUi
    class Config(BaseUiLinker.Config):
        default_widget_type: str = ""
        widget_type_list: Optional[List[str]] = None 


    _linker = None
    layout_linker = None
    
    class Data(BaseModel):
        curent_widget_type:  str = ""
               
    def update(self, data):
        pass
    
    def connect(self, downloader, obj, data = None):
        #if data is None:
        #    data = self.Data()
        self.layout_linker = LayoutLinker(self.widget.device_layout)
        self.layout_linker.connect(downloader)
        #self.setup_ui(obj, data)
        super().connect(downloader, obj, data)
            
    def setup_ui(self, device, data):
        if self.config.widget_type_list is None:
            self.config.widget_type_list =  get_widget_types(device.config.type)
        if not self.config.widget_type_list:
            raise ValueError(f"No Widget found for device of type {device.config.type}")
        if not self.config.default_widget_type:
            self.config.default_widget_type = self.config.widget_type_list[0]
                
        self.widget.style_switch.clear()    
        for wtype in self.config.widget_type_list:
            self.widget.style_switch.addItem(wtype)
        self.widget.style_switch.setCurrentIndex(0)
        
        
        def on_switch_changed(wt):
            # clear the layout
            self.layout_linker.clear()
            self.layout_linker.add_device(device, wt)
            data.curent_widget_type = wt
            
        
        self.actions.add(
            on_switch_changed, 
            [self.widget.style_switch.currentText], 
        ).connect_combo(self.widget.style_switch)
        
        # load with the first widget on the list or default 
        if self.config.default_widget_type and self.config.default_widget_type in self.config.widget_type_list:        
            on_switch_changed(self.config.default_widget_type)
        else:
            on_switch_changed(self.config.widget_type_list[0])
        
            
            
                



""" The idea here is to create interface between a data value and a widget 



Three type of interfaces :
    - input : The widget is a text, checkbox, button, etc .. and the data is an input of the application
            It has the methods:
                 set_input(val) :  update the input widget with given value
                 get_input()    :  retrieve the value with the right data type
                 
    - output: The widget is a representation of a value, e.g. QLabel the data is an output of the application 
            It has the methods:
                set_output(val) : update the output widget (e.g. QLabel) with the value of dtype
     
    - input output: Two widgets represent input and output if they differ the output is hilighted in red 
                    This is to be used in a config pannel for instance
            t has the methods:
                 set_input(val)   
                 set_output(val) 
                 get_input()  
                 
    Each of the interface has a variable number of argument and keyword depending on the nature of 
    the data value. 
    For performance purpose The update_* methods are created at frontend in the __init__ 
    function to widgets types. This is to avoid things like isinstance(widget, QLabel) at run time 
    The init shall handle how to update and get widget. So far only used case are set, one can increase 
    the diversity of used widget used.
"""

from collections import OrderedDict
from pydantic import BaseModel
from typing import Callable

def output(cond, func=None):
    if func is None:
        return Output(cond).funcsetter
    else:
        return Output(cond, func)
        
class Output:
    def __init__(self, cond, func=None):
        self.func = func
        self.cond = cond
    
    def funcsetter(self, func):
        self.func = func
        return self
    
    def build(self, output_widget, config):
        return self.func(output_widget, config)
    
    def check(self, output_widget):
        return self.cond(output_widget)    
            
class OutputType(type):
    def __new__(cls, name, bases, d):
        
        _outputs = []
        
        for k,v in d.items():
            if isinstance(v, Output):
                _outputs.append(v) 
        d ['__outputs__'] = _outputs        
        return type.__new__(cls, name, bases, d)
    
class OutputVal(metaclass=OutputType):
    class Config(BaseModel):
        pass
        
    def __init__(self, output_widget, config=None, **kwargs):        
        self.config = self.parse_config(config, **kwargs)
        for fout in self.__outputs__:
            if fout.check(output_widget):
                set = fout.build( output_widget, self.config)
                break
        else:
            raise ValueError(f"Cannot found suitable output for class {self.__class__.__name__} and widget {type(output_widget)}")
        
        self.set = set 
        
            
        
    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            return cls.Config( **{**__config__, **kwargs} )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")


class BoolVal_O(OutputVal):
    class Config(OutputVal.Config):
        fmt: Callable  = staticmethod(lambda v: "[X]" if v else "[ ]")
    
    @output(lambda w: hasattr(w, "setText"))
    def _text_output_builder(w, c):
        def set_output(b):
            t = c.fmt(b)                
            w.setText(t)     
        return set_output
    
            
    
    
        



from ._core_base import (_BaseObject, _BaseProperty, BaseData, open_object, 
                        ChildrenCapability, ChildrenCapabilityConfig
                        )
from ._class_recorder import  KINDS,  record_class
from ._core_node import BaseNode 
from ._core_interface import BaseInterface
from ._core_rpc import BaseRpc

from ._core_obj_dict import ObjDict


from typing import  Optional, Any 





class BaseDeviceConfig(_BaseObject.Config, ChildrenCapabilityConfig):
    kind: KINDS = KINDS.DEVICE.value
    type: str = "Base"
    
    
    def cfgdict(self, exclude=set()):
        all_exclude = {*{}, *exclude}
        d = super().cfgdict(exclude=all_exclude)       
        return d
    
  
    
def open_device(cfgfile, path=None, prefix="", key=None, default_type=None, **kwargs):
    """ Open a device from a configuration file 

        
        Args:
            cfgfile: relative path to one of the $CFGPATH or absolute path to the yaml config file 
            key: Key of the created Manager 
            path (str, int, optional): 'a.b.c' will loock to cfg['a']['b']['c'] in the file. If int it will loock to the Nth
                                        element in the file
            prefix (str, optional): additional prefix added to the name or key

        Output:
            device (BaseDevice subclass) :tanciated Device class     
    """
    kwargs.setdefault("kind", KINDS.DEVICE)

    return open_object(cfgfile, path=path, prefix=prefix, key=key, default_type=default_type, **kwargs) 




class DeviceProperty(_BaseProperty):    
    fbuild = None    
    
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self
     
   
    def __call__(self, func):
        """ The call is used has fbuild decorator 
        
        this allows to do
        
        ::
            
            class MyManager(BaseManager):
                @MyDevice.prop('motor2')
                def motor2(self, motor):
                    # do somethig
                    
        """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, device):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, device)  



class DeviceDict(ObjDict):
    class Config(ObjDict.Config):
        kinds = set([KINDS.DEVICE])

        default_kind = KINDS.DEVICE

@record_class
class BaseDevice(_BaseObject, ChildrenCapability):
    Property = DeviceProperty
    Config = BaseDeviceConfig
    Interface = BaseInterface
    Data = BaseData
    Node = BaseNode
    Rpc = BaseRpc    
    Dict = DeviceDict 
    
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] = None,
           **kwargs
        ) -> None:        
        
        super().__init__(key, config=config, **kwargs)  
        if self._localdata is None:
            self._localdata = {}
    

    @classmethod
    def parse_config(cls, config, **kwargs):
        if isinstance(config, dict):
            kwargs = {**config, **kwargs}
            config = None
        return super().parse_config( config, **kwargs)
        
    @classmethod
    def new_com(cls, config: Config, com: Optional[Any] = None) -> Any:
        """ Create a new communication object for the device 
            
        Args:
           config: Config object of the Device Class to build a new com 
           com : optional, A parent com object used to build a new com if applicable  
           
        Return:
           com (Any): Any suitable communication object  
        """
        return com 
    
        
    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()  
            
            
    def connect(self):
        """ Connect device to client """
        raise NotImplementedError('connect method not implemented') 
    
    def disconnect(self):
        """ Disconnect device from client """
        raise NotImplementedError('disconnect method not implemented')    
    
    def is_connected(self):
        """ True if device connected """
        raise NotImplementedError('is_connected method not implemented') 
    
    def rebuild(self):
        """ rebuild will disconnect the device and create a new com """
        self.disconnect()
        self.clear()
        self._com = self.new_com(self._config)
    
        




from .base import BaseUiLinker, record_widget_factory, get_widget_types, get_widget_factory
from .widget_io import Inputs, Outputs, InputOutputs
from .style import STYLE_DEF, STYLE, get_style
from .io import find_ui
from .manager import ManagerLinker



""" The idea here is to create interface between a data value and a widget 



Three type of interfaces :
    - input : The widget is a text, checkbox, button, etc .. and the data is an input of the application
            It has the methods:
                 set_input(val) :  update the input widget with given value
                 get_input()    :  retrieve the value with the right data type
                 
    - output: The widget is a representation of a value, e.g. QLabel the data is an output of the application 
            It has the methods:
                set_output(val) : update the output widget (e.g. QLabel) with the value of dtype
     
    - input output: Two widgets represent input and output if they differ the output is hilighted in red 
                    This is to be used in a config pannel for instance
            t has the methods:
                 set_input(val)   
                 set_output(val) 
                 get_input()  
                 
    Each of the interface has a variable number of argument and keyword depending on the nature of 
    the data value. 
    For performance purpose The update_* methods are created at frontend in the __init__ 
    function to widgets types. This is to avoid things like isinstance(widget, QLabel) at run time 
    The init shall handle how to update and get widget. So far only used case are set, one can increase 
    the diversity of used widget used.
"""
from PyQt5.QtWidgets import QLineEdit, QLabel, QCheckBox
from .style import get_style, STYLE
from .style import STYLE
from typing import Tuple, Any
import math

FLOAT_FMT = "%.3f"
FLOAT_DEFAULT = 0.0
INT_FMT = "%d"
INT_DEFAULT = 0
STR_FMT = "%s"
STR_DEFAULT = ""




def BOOL_FMT(b):
    """ CheckBox string representation"""
    return "[X]" if b else "[_]"
BOOL_DEFAULT = False

ENUM_FMT = "%s: %s"


def float_nan(t):
    """ return float(value) or None if value is a Nan """
    try:
        return float(t)
    except (ValueError, TypeError):
        return None

def int_nan(t):
    """ return int(value) or None if value is a Nan """
    try:
        return int(t)
    except (ValueError, TypeError):
        return None    


def switch_style(w,b):
    style = STYLE.SIMILAR if b else STYLE.DIFFERENT
    w.setStyleSheet(get_style(style))


def gattr(key):
    return lambda data: getattr(data, key)
    
def sattr(key):
    return lambda data, value: setattr(data, key, value)

def gsattr(key):
    return lambda data: getattr(data, key), lambda data, value: setattr(data, key, value)


def _dummy(data):
    pass

class BaseVal_IO:
    def __init__(self, set_input, set_output, get_input):
                        
        self.set_input  = set_input
        self.set = set_output
        self.get  = get_input        
        
    def disconnect(self):        
        self.set_input =  _dummy
        self.set = _dummy
        self.get = _dummy

class BaseVal_I:
    def __init__(self, set_input, get_input):
        self.set_input  = set_input        
        self.get  = get_input
    
    def set(self, data):
        raise ValueError("No widget output associated")
        
    def disconnect(self):        
        self.set_input =  _dummy        
        self.get = _dummy        

class BaseVal_O:
    def __init__(self, set_output):             
        self.set = set_output
    
    def set_input(self, data):
        raise ValueError("No widget input associated")
        
    def disconnect(self):          
        self.set = _dummy

class BoolVal_IO(BaseVal_IO):
    def __init__(self, input, output,  default=False, fmt=BOOL_FMT):            
        if hasattr(input, "setChecked") and hasattr(input, "isChecked"):
            def set_input(b):                
                input.setChecked(b)
            def get_input():
                return input.isChecked()                
        else:
            raise ValueError("expecting a QCheckBox as input  %s"%(type(input)))
        
        if hasattr(output, "setChecked"):
            def set_output(b):                           
                output.setChecked(b)
                switch_style(output, input.isChecked()==b)       
        
        elif  hasattr(output, "setText"):
            def set_output(b):
                t = fmt(b)                
                output.setText(t)
                switch_style(output, input.isChecked()==b)                
        else:
            raise ValueError("expecting a QLabel or QCheckbox as output got %s"%(type(output)))
                
        super().__init__(set_input, set_output, get_input)    
        if default is not None:        
            input.setChecked(default)

class BoolVal_O(BaseVal_O):
    def __init__(self, output, fmt=BOOL_FMT):   
        if hasattr(output, "setChecked"):
            def set_output(b):
                output.setChecked(b)
                switch_style(output, input.isChecked()==b)         
        elif hasattr(output, "setText"):
            def set_output(b):
                t = fmt(b)                
                output.setText(t)        
        else:
            raise ValueError("expecting a QCheckBox as input and QLabel as output got %s"%(type(output)))    
        super().__init__( set_output)

class BoolVal_I(BaseVal_I):    
    def __init__(self, input,  default=None):            
        if isinstance(input, QCheckBox):
            def set_input(b):                
                input.setChecked(b)
            
            def get_input():
                return input.isChecked()                
        else:
            raise ValueError("expecting a QCheckBox as input got %s"%(type(input)))    
        super().__init__(set_input, get_input)            
        if default is not None:
            input.setChecked(bool(default))


class NBoolVal_IO(BaseVal_IO):
    def __init__(self, input, output, default=False, fmt=BOOL_FMT):            
        if hasattr(input, "setChecked") and hasattr(input, "isChecked"):
            def set_input(b):                
                input.setChecked(not b)
            def get_input():
                return not input.isChecked()                
        else:
            raise ValueError("expecting a QCheckBox as input  %s, %s"%(type(input)))
        
        if hasattr(output, "setChecked"):
            def set_output(b):                           
                output.setChecked(not b)
                switch_style(output, input.isChecked()==(not b))       
        
        elif  hasattr(output, "setText"):
            def set_output(b):
                t = fmt(not b)                
                output.setText(t)
                switch_style(output, input.isChecked()==(not b))                
        else:
            raise ValueError("expecting a QLabel as output got %s, %s"%(type(output)))
                
        super().__init__(set_input, set_output, get_input)    
        if default is not None:        
            input.setChecked(bool(default))

class NBoolVal_O(BaseVal_O):
    def __init__(self, output, fmt=BOOL_FMT):   
        if hasattr(output, "setChecked"):
            def set_output(b):
                output.setChecked(not b)
                     
        elif hasattr(output, "setText"):
            def set_output(b):
                t = fmt(not b)                
                output.setText(t)        
        else:
            raise ValueError("expecting a QCheckBox as input and QLabel as output got %s"%(type(output)))    
        super().__init__(set_output)

class NBoolVal_I(BaseVal_I):    
    def __init__(self, input, default=False):            
        if isinstance(input, QCheckBox):
            def set_input(b):                
                input.setChecked(not b)
            
            def get_input():
                return not input.isChecked()                
        else:
            raise ValueError("expecting a QCheckBox as input and QLabel as output got %s"%(type(input)))    
        super().__init__(set_input, get_input)            
        input.setChecked(default)


class StrVal_IO(BaseVal_IO):
    def __init__(self, input, output,  fmt=STR_FMT, default=STR_DEFAULT):
        if isinstance(fmt, str):
            fmt = lambda x, fmt=fmt: fmt%x
        
        
        if hasattr(input, "setText") and hasattr(output, "setText"):
            def set_input(v):                    
                input.setText(fmt(v))
            def set_output(v):                
                t = fmt(v)                
                output.setText(t)
                switch_style(output, input.text()==t)
            def get_input():
                return input.text()                
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__( set_input,set_output,get_input)    
        if default is not None:
            input.setText(fmt(default))

class StrVal_I(BaseVal_I):
    def __init__(self, input, feedback=None, fmt=STR_FMT, default=STR_DEFAULT):    
        if isinstance(fmt, str):
            fmt = lambda x, fmt=fmt: fmt%x
            
        
        if hasattr(input, "setText"):
            def set_input(v):                   
                input.setText(fmt(v))            
            def get_input():
                return input.text()                
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input, get_input)    
        if default is not None:
            input.setText(fmt(default))
            

class StrVal_O(BaseVal_O):
    def __init__(self, output,  fmt=STR_FMT):
        if isinstance(fmt, str):
            fmt = lambda x,fmt=fmt: fmt%x
                
        if hasattr(output, "setText"):            
            def set_output(v):
                t = fmt(v)                
                output.setText(t)                            
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_output)        


class FloatVal_IO(BaseVal_IO):
    def __init__(self, input, output, feedback=None, fmt=FLOAT_FMT, default=FLOAT_DEFAULT):
        if isinstance(fmt, str):
            fmt = lambda x, fmt=fmt: fmt%x
        
        
        if hasattr(input, "setText") and hasattr(output, "setText"):
            def set_input(v):                        
                input.setText(fmt(v))
            
            def set_output(v):                
                t = fmt(v)                
                output.setText(t)
                switch_style(output, float_nan(input.text())==v)
                
            def get_input():
                try:
                    return float(input.text())  
                except ( TypeError, ValueError):
                    return math.nan

        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input, set_output, get_input)    
        if default is not None:
            input.setText(fmt(default))

class FloatVal_I(BaseVal_I):
    def __init__(self, input, fmt=FLOAT_FMT, default=FLOAT_DEFAULT):    
        if isinstance(fmt, str):
            fmt = lambda x, fmt=fmt: fmt%x
            
        
        if hasattr(input, "setText"):
            def set_input(v):                           
                input.setText(fmt(v))            
            def get_input():
                try:
                    return float(input.text())  
                except ( TypeError, ValueError):
                    return math.nan
                                    
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input,get_input)    
        if default is not None:
            input.setText(fmt(default))
            

class FloatVal_O(BaseVal_O):
    def __init__(self, output,  fmt=FLOAT_FMT):
        if isinstance(fmt, str):
            fmt = lambda x,fmt=fmt: fmt%x
                
        if hasattr(output, "setText"):            
            def set_output(v):                
                t = fmt(v)                
                output.setText(t)                            
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_output)        

class IntVal_IO(BaseVal_IO):
    def __init__(self,  input, output,fmt=INT_FMT, default=INT_DEFAULT):
        if isinstance(fmt, str):
            fmt = lambda x,fmt=fmt: fmt%x
                
        
        if hasattr(input, "setText") and hasattr(output, "setText"):
            def set_input(v):                            
                input.setText(fmt(v))
            def set_output(v):
                t = fmt(v)               
                output.setText(t)
                switch_style(output, int_nan(input.text())==v)
            def get_input():
                return int(input.text())                
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input, set_output, get_input)    
        if default is not None:
            input.setText(fmt(default))

class IntVal_I(BaseVal_I):
    def __init__(self, input,  fmt=INT_FMT, default=INT_DEFAULT):
        if isinstance(fmt, str):
            fmt = lambda x,fmt=fmt: fmt%x
                
        
        if hasattr(input, "setText"):
            def set_input(v):
                input.setText(fmt(v))            
            def get_input():
                return int(input.text())                
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input,get_input)    
        if default is not None:
            input.setText(fmt(default))

class IntVal_O(BaseVal_O):
    def __init__(self,  output,  fmt=INT_FMT):
        if isinstance(fmt, str):
            fmt = lambda x,fmt=fmt: fmt%x
                        
        if hasattr(output, "setText"):            
            def set_output(v):                
                t = fmt(v)               
                output.setText(t)            
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_output)        

        
class EnumVal_IO(BaseVal_IO):    
    def __init__(self,  enum, input, output,  fmt="%s: %s", default=None):
    
        num_to_index = {e.value:i for i,e in enumerate(enum)}
        index_to_num = {i:e.value for i,e in enumerate(enum)}
        
        if isinstance(fmt, str):
            fmt = lambda x,y,fmt=fmt: fmt%(x,y)
            
        if hasattr(input, "setCurrentIndex") and hasattr(output, "setText"):
            def set_input(v):                            
                input.setCurrentIndex( num_to_index[v] )
                
            def set_output(a):
                a =  enum(a)                                                
                output.setText( fmt(a.value, a.name))             
                switch_style(output, input.currentIndex()==num_to_index[a.value])
                
            def get_input():
                i = input.currentIndex()
                return index_to_num[i]                
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__( set_input, set_output, get_input)  
        input.clear()
        input.addItems( [a.name for a in enum] )
        if default is not None:
            input.setCurrentIndex( num_to_index[default] )    

class EnumVal_I(BaseVal_I):
    
    def __init__(self, enum, input, default=None):
        
        num_to_index = {e.value:i for i,e in enumerate(enum)}
        index_to_num = {i:e.value for i,e in enumerate(enum)}
        if hasattr(input, "setCurrentIndex"):
            def set_input(v):                             
                input.setCurrentIndex( num_to_index[v] )
                            
            def get_input():
                i = input.currentIndex()                
                return  index_to_num[i]
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_input,get_input)  
        input.clear()
        input.addItems( [a.name for a in enum] )
        if default is not None:
            input.setCurrentIndex( num_to_index[default] )    

        
class EnumVal_O(BaseVal_O):   
    def __init__(self, enum, output, fmt="%s: %s"):
        num_to_index = {e.value:i for i,e in enumerate(enum)}
        
        if isinstance(fmt, str):
            fmt = lambda x,y,fmt=fmt: fmt%(x,y)
        
        if hasattr(output, "setText"):                            
            def set_output(a):
                a =  enum(a)                                                
                output.setText( fmt(a.value, a.name))               
                switch_style(output, input.currentIndex()==num_to_index[a.value])                            
        else:
            raise ValueError("Invalid input output combination")    
        super().__init__(set_output)     


class Feedback_O(BaseVal_O):
    """ A feedback widget takes a couple of (er, msg) 
    
    if er the text msg is screened with an STYLE.ERROR  style 
    else  msg is screened with an STYLE.NORMAL  style 
    """
    def __init__(self, output):                
        if hasattr(output, "setText") and hasattr(output, "setStyleSheet") :                            
            def set_output(er_msg: Tuple[Any,str]) -> None:
                er,msg = er_msg
                if er:
                    output.setStyleSheet(get_style(STYLE.ERROR))    
                else:
                    output.setStyleSheet(get_style(STYLE.NORMAL))                        
                output.setText(msg)                                   
        else:
            raise ValueError("Invalid input output combination")            
        super().__init__(set_output)
        


class Inputs:
    Base = BaseVal_I
    Bool = BoolVal_I
    NBool = NBoolVal_I
    Str = StrVal_I
    Float = FloatVal_I
    Int = IntVal_I
    Enum = EnumVal_I

class Outputs:
    Base = BaseVal_O
    Bool = BoolVal_O
    NBool = NBoolVal_O
    Str = StrVal_O
    Float = FloatVal_O
    Int = IntVal_O
    Enum = EnumVal_O  
    Feedback = Feedback_O  

class InputOutputs:
    Base = BaseVal_IO
    Bool = BoolVal_IO
    NBool = NBoolVal_IO
    Str = StrVal_IO
    Float = FloatVal_IO
    Int = IntVal_IO
    Enum = EnumVal_IO    




        



from ._core_objects import  BaseNode, NodeAlias, NodeAlias1, record_class, parser

from collections import deque
import time
from  datetime import datetime, timedelta
from typing import Union, List, Dict, Optional
from pydantic import validator
from enum import Enum
from py_expression_eval import Parser
from dataclasses import dataclass
import math 


_eval_parser = Parser()


__all__ = [
"UnixTimeNode",
"LocalTimeNode",
"LocalUtcNode", 
"DequeNode", 
"DequeNode1",
"local_utc", 
"local_utc2", 
"local_time",
"AllTrue", 
"AllFalse", 
"AnyTrue", 
"AnyFalse",
"NegNode", 
"InsideIntervalNode", 
"InsideCircleNode", 
"PosNameNode", 
"FormulaNode", 
"FormulaNode1",
"PolynomNode1",
"StatisticNode", 
"CounterNode", 
"FormatNode", 
"SumNode", 
"MeanNode", 
"MinNode", 
"MaxNode"
]

@record_class
class UnixTimeNode(BaseNode):
    """ A basic node returning the float local time  
    
    Args:
        key (optional, str): node key
        delta (optional, float): time shift in seconds 
    
    """
    class Config(BaseNode.Config):
        type = "UnixTime"
        delta: float = 0.0
    
    def __init__(self, key: str = 'local_unix_time', config=None, **kwargs):
        super().__init__(key, config=config, **kwargs)
        
    def fget(self) -> float:
        return time.time()+self.config.delta 

@record_class
class LocalTimeNode(BaseNode):
    """ A basic node returning the float local time  
    
    Args:
        key (optional, str): node key
        delta (optional, float): time shift in seconds 
        
    Properties:
        _delta : the _delta timeshift in second 
    """
    class Config(BaseNode.Config):
        type = "LocalTime"
        delta: float = 0.0 # delta in seconds 
        
    def __init__(self, key: str = 'local_time', config=None, **kwargs):
        super().__init__(key, config=config, **kwargs)
    
    def fget(self) -> float:
        return datetime.now()+timedelta(seconds=self.config.delta)
        

local_time = LocalTimeNode('local_time')

@record_class
class LocalUtcNode(BaseNode):
    """ A basic node returning the local UTC as string 
    
    Args:
        key (str, optional): node key
        delta (float, optional): time shift in seconds 
        format (str, optional): Returned format, default is iso 8601 '%Y-%m-%dT%H:%M:%S.%f%z'
    Properties:
        _delta : the _delta timeshift in second 
    """
    class Config(BaseNode.Config):
        type = "LocalUtc"
        delta: float = 0.0
        format: str = '%Y-%m-%dT%H:%M:%S.%f%z'
        
    def __init__(self, key : str ='local_utc', config=None, **kwargs):
        super().__init__(key, config=config, **kwargs)
                
    def fget(self) -> str:
        tc = datetime.utcnow()+timedelta(seconds=self.config.delta)   
        return tc.strftime(self.config.format)
        
##
# Add an instance to the local_utc 
local_utc  = LocalUtcNode('local_utc')
local_utc2 = LocalUtcNode('local_utc', format='%Y-%m-%d-%H:%M:%S.%f%z')


@record_class
class DequeNode(NodeAlias):
    """ This is an :class:`NodeAlias` returning at each get a :class:`collections.deque` 
    
    Specially handy for live plot and scopes
    
    Args:
       key (string): alias node keyword 
       nodes (list of :class:`UaNode`,:class:`NodeAlias`): 
              List of nodes to get()  
       maxlen (int): maximum len of the dequeu object  
    
    Example:
        
    In this example the first motor is moving while the second standstill at 0.0
         
    ::
    
        >>> from pydevmgr_core import local_time
        >>> time = UnixTimeNode('time') 
        >>> nodes = [time, mgr.motor1.stat.pos_actual, mgr.motor2.stat.pos_actual]
        >>> f = DequeNode( "mot_poses", nodes, 20)
        >>> f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0)])
        >>> f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0),
        (1604931614.258859, 95.32508185106808, 0.0)])
        >>> f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0),
        (1604931614.258859, 95.32508185106808, 0.0),
        (1604931615.451603, 94.13508185106821, 0.0)])
        >>> etc ....
        
    """
    class Config(NodeAlias.Config):
        type = "Deque"
        maxlen: int = 10000
        trigger_index: Optional[int] = None
    
    def __init__(self, 
          key: Optional[str] = None, 
          nodes: Optional[Union[BaseNode,List[BaseNode]]] = None, 
          config: Optional[Config] = None,
          **kwargs
        ) -> None: 
        
        if nodes is None:
            raise ValueError("nodes cannot be None")    
        super().__init__(key, nodes, config, **kwargs)        
        
        self._data = deque([], self._config.maxlen)
        self._scalar = not hasattr(nodes, "__iter__")
                
        
    @property
    def data(self):
        return self._data 
    
    @property
    def columns(self):
        return [n.key for n in self.nodes]
    
        
    def fget(self, *values): 
        if self.config.trigger_index is not None and not values[self.config.trigger_index]:
            return self._data      
        
        if self._scalar:
            self._data.append(values[0])
        else:
            self._data.append(values)
        return self._data
    
    def reset(self, maxlen=None):
        if maxlen is not None:
            self.config.maxlen = maxlen
                
        if self._maxlen != self._config.maxlen:             
            # maxlen has been changed 
            self._maxlen = self._config.maxlen        
            self._data = deque([], self._config.maxlen)
        else:
            self._data.clear()
        
@record_class
class DequeNode1(NodeAlias1):
    """ This is an :class:`NodeAlias1` returning at each get a :class:`collections.deque` 
    
    Specially handy for live plot and scopes
    
    Args:
       key (string): alias node keyword 
       node  ( :class:`UaNode`,:class:`NodeAlias`): 
              
       maxlen (int): maximum len of the dequeu object  
    
    .. seealso::
       :class:`DequeNode` 
        
    """
    class Config(NodeAlias.Config):
        type = "Deque1"
        maxlen: int = 10000        
    
    def __init__(self, 
          key: Optional[str] = None, 
          node: Optional[Union[BaseNode,List[BaseNode]]] = None, 
          config: Optional[Config] = None,
          **kwargs
        ) -> None: 
        
        if node is None:
            raise ValueError("node cannot be None")    
        super().__init__(key, node, config, **kwargs)
        self._maxlen = self._config.maxlen        
        self._data = deque([], self._config.maxlen)                        
        
    @property
    def data(self):
        return self._data 
        
    def fget(self, value):         
        self._data.append(value)
        return self._data
    
    def reset(self, maxlen=None):
        if maxlen is not None:
            self.config.maxlen = maxlen
                
        if self._maxlen != self._config.maxlen:             
            # maxlen has been changed 
            self._maxlen = self._config.maxlen        
            self._data = deque([], self._config.maxlen)
        else:
            self._data.clear()


@record_class
class AllTrue(NodeAlias):
    class Config(NodeAlias.Config):
        type = "AllTrue"
    def fget(*nodes):
        return all(nodes)

@record_class
class AnyTrue(NodeAlias):
    class Config(NodeAlias.Config):
        type = "AnyTrue"
    def fget(*nodes):
        return any(nodes)
        
@record_class        
class AllFalse(NodeAlias):
    class Config(NodeAlias.Config):
        type = "AllFalse"
    def fget(*nodes):
        return not any(nodes)

@record_class
class AnyFalse(NodeAlias):
    class Config(NodeAlias.Config):
        type = "AnyFalse"
    def fget(*nodes):
        return not all(nodes)

@record_class
class NegNode(NodeAlias1):
    """ Return the boolean oposite of the aliased node """
    class Config(NodeAlias1.Config):
        type = "NegNode"
    def fget(self, value):
        return not value 

@record_class
class CounterNode(BaseNode):
    """ A simple counter node at each get the counter is increased and returned 
    
    Args:
        start (optional, int): start number 
        
    Example:
    
    ::
    
       >>> c = CounterNode()
       >>> c.get()
       1
       >>> c.get()
       2
       >>> c.reset()
       >>> c.get()
       1
    """
    class Config(BaseNode.Config):
        type: str = "Counter"
        start: int  = 0
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
    
    def reset(self):
        self._counter = self.config.start
    
    def fget(self):
        self._counter += 1
        return self._counter 
        
@record_class
class InsideIntervalNode(NodeAlias1):
    """ Bool Node alias to check if a value is inside a given interval 
    
    Args:
        key (str):  node key
        node (:class:`BaseNode`): node returning a float 
        min (float, optional): min value of the interval 
        max (float, optional): max value of the interval
    """
    class Config(NodeAlias1.Config):
        type = "InsideInterval"
        min : Optional[float] = None
        max : Optional[float] = None
            
    def fget(self, value):
        c = self.config
        if c.min is not None and value<c.min:
            return False
        if c.max is not None and value>c.max:
            return False
        return True    
        
@record_class
class InsideCircleNode(NodeAlias):
    """ Bool Node alias to check if a 2d position is inside a circle 
    
    Args:
        key (str): node key
        nodes (list of :class:`BaseNode`): two nodes returning x and y coordinates 
        x0, y0 (float): circle origin  default is (0.0, 0.0)
        r (float): circle radius (default is 1.0)
    
    """
    class Config(NodeAlias.Config):
        type = "InsideCircle"
        x0 : float = 0.0
        y0 : float = 0.0 
        r  : float = 1.0
            
    _n_nodes_required = 2    
    
    def fget(self, x, y):
        c = self.config
        return ((x-c.x0)**2 + (y-c.y0)**2) < (c.r*c.r)        
        

@record_class    
class PosNameNode(NodeAlias1):
    """ Node alias returning a position name thanks to a list of position and a tolerance 
    
    Args:
        key (str):  node key 
        node (:class:`BaseNode`)
        poses (dict):  name/position pairs
        tol (float): tolerence for each poses
        unknown (str, optional): string for unknown position default is ""
        
    Example:
       
    :: 
       
       PosNameNode('pos_name', motor1.stat.pos_actual, {'FREE':0.0, ''})
    """
    class Config(NodeAlias1.Config):
        type = "PosName"
        poses: Dict[str,float] = None
        tol: float = 0.0 
        unknown: str = ""
        
    def fget(self, value: float) -> str:
        c = self.config
        for name, pos in c.poses.items():
            if abs(pos-value)<c.tol:
                return name
        return c.unknown
         
     
@record_class    
class FormulaNode(NodeAlias):
    class Config(NodeAlias.Config):
        type: str = "Formula"
        formula : str = "-99.99"
        varnames: Optional[Union[List[str],str]] = None  
    
    def __init__(self, *args, **kwargs):                            
        super().__init__(*args, **kwargs)
        if isinstance(self.config.varnames, str):
            varnames = [self.config.varnames]
        else:
            varnames = self.config.varnames
        if varnames is  None:
            if len(self.nodes)==1:
                varnames = ('x',)
            else:
                varnames = tuple( 'x'+str(i+1) for i in range(len(self.nodes)) )
        elif len(varnames)!=len(self.nodes):
            raise ValueError( f"Got {len(self.nodes)} nodes but configured varnames has size {len(self.config.varnames)}")
                    
        self._parser = _eval_parser.parse(self.config.formula)
        self._inputs = {'nan': math.nan}
        
        self._varnames = varnames
            
                
    def fget(self, *values):   
        v = {k:x for k,x in zip(self._varnames, values)}
        return self._parser.evaluate(v)


@record_class    
class FormulaNode1(NodeAlias1):
    class Config(NodeAlias1.Config):
        type: str = "Formula1"
        formula : str = "-99.99"
        varname: str = 'x' 
    
    def __init__(self, *args, **kwargs):                            
        super().__init__(*args, **kwargs)
                
        self._parser = _eval_parser.parse(self.config.formula)
        self._inputs = {'nan': math.nan}
        
        self._varname = self.config.varname
                            
    def fget(self, value):          
        return self._parser.evaluate({self._varname:value})


@record_class
class PolynomNode1(NodeAlias1):
    """ Apply a polynome 

    Config:
        polynom (list) : list of coefficients (lowest first) 
    """
    class Config(NodeAlias1.Config):
        polynom: List[float] = [0.0,1.0]
        
    
    def fget(self, value):
        if not self.config.polynom:
            return 0.0
        x = self.config.polynom[0]
        for i,c in enumerate(self.config.polynom[1:], start=1):
            x += c*value**i
        return x

PolynomNode = PolynomNode1 



@record_class
class StatisticNode(NodeAlias1):
    class Config(NodeAlias1.Config):
        mean: float = 0.0 # expected mean for Variance and rms computation
        type: str = "Statistic"
        
    @dataclass
    class Stat:
        min  : float = -math.inf
        max  : float =  math.inf
        sum  : float =  math.nan 
        mean : float =  math.nan
        sum2 : float =  math.nan
        rms  : float =  math.nan 
        n    : int = 0
      
    def __init__(self, *args,  **kwargs):                            
        super().__init__(*args, **kwargs)
        self._stat = self.Stat()
        self.reset()
        
    def reset(self):
        s = self._stat
        s.min = -math.inf
        s.max =  math.inf
        s.sum =  math.nan 
        s.mean = math.nan
        s.sum2 = math.nan
        s.rms =  math.nan 
        s.n = 0
        
    def fget(self, value):
        s= self._stat
        if not s.n:
            s.min = value 
            s.max = value 
            s.sum = value
            s.mean = value 
            s.sum2 = (value - self.config.mean)**2
            s.var = s.sum2
            s.rms = value 
            s.n = 1
            
        else:
            s.sum = s.sum+value
            s.sum2 = s.sum2 + (value - self.config.mean)**2
            s.n = s.n+ 1
            s.var = s.sum2/s.n
            s.min = min(value, s.min)
            s.max = max(value, s.max)                         
            s.mean = s.sum/s.n
            s.rms = math.sqrt(s.var) 
            
        return s

class _Stat(NodeAlias1):
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
    

@record_class
class SumNode(_Stat):
    class Config(NodeAlias1.Config):        
        type: str = "Sum"
        
    def reset(self):
        self._sum = 0.0
        
    def fget(self, value):
        self._sum += value 
        return self._sum    
        
@record_class
class MeanNode(_Stat):
    class Config(NodeAlias1.Config):        
        type: str = "Mean"
        
    def reset(self):
        self._sum = 0.0
        self._n = 0
    
    def fget(self, value):
        self._sum += value 
        self._n += 1
        return self._sum / self._n

@record_class
class MinNode(_Stat):
    class Config(NodeAlias1.Config):        
        type: str = "Min"
    
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
        
    def reset(self):
        self._min = math.inf
            
    def fget(self, value):
        if value < self._min:            
            self._min = value
        return self._min

@record_class
class MaxNode(_Stat):
    class Config(NodeAlias1.Config):        
        type: str = "Max"
    
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
        
    def reset(self):
        self._max = -math.inf
            
    def fget(self, value):
        if value > self._max:            
            self._max = value
        return self._max

        
@record_class
class FormatNode(NodeAlias):
    class Config(NodeAlias.Config):
        type: str = "Format"
        format: str = "{0}"
    
    def fget(self, *values):
        return self.config.format.format(*values)
        
        



from .widget_io import Inputs, Outputs, InputOutputs
from .io import find_ui
from pydevmgr_core import Downloader, _BaseObject, DataLink, BaseData, RpcError
from typing import Union, Optional, Callable, Iterable
from PyQt5.QtWidgets import QLayout,  QBoxLayout,  QGridLayout , QWidget, QAction, QComboBox, QVBoxLayout, QBoxLayout, QGridLayout
from PyQt5 import uic, QtCore
from pydantic import BaseModel
#QCheckBox, QAction, QPushButton, QComboBox, QLayout,  QWidget, QFrame

class Action:
    """ An action object must be created from a pool of actions of class :class:`Actions` 
    
    Args:
       - parent: :class:`Actions`
       - func (Callable): Command function must take len(inputs) arguments
       - inputs (List): list of static or callable arguments for func
                        if callable it will be called at run time 
                        otherwhise they are static args, sent has is 
       - feedback (Callable) : function of signature f(er, msg) called for feedback. 
                            In case of action this is called only on exception 
        - before (Callable): function without argument called before ``func`` 
        - after (Callable): function without argument called after ``func`` 
    """
    def __init__(self, parent, 
          func: Callable, 
          inputs: list = None, 
          feedback: Callable = None, 
          before: Callable = None, 
          after: Callable = None, 
          okay_msg: str = ""
        ) -> None:        
        inputs = [] if inputs is None else inputs        
        # transform contant to lambdas function
        cleaned_inputs = []
        
        self._disconnection = parent._disconnection
        self._combos = parent._combos
                        
        # everything not callable is treated as a constant 
        # create a dummy function to return this constant
        # better to do it frontend rather then during execution                 
        for input in inputs:
            if not hasattr(input, "__call__"):
                input = lambda __cte__ = input: __cte__
            cleaned_inputs.append(input)
                
        
        if feedback is None:
            def func_call():   
                if before: before()   
                func(*(fi() for fi in cleaned_inputs))
                if after: after()
        else:         
            def func_call():            
                try:
                    if before: before()   
                    func(*(fi() for fi in cleaned_inputs))                    
                except (RpcError,TypeError,ValueError,RuntimeError) as er:
                    feedback(er, str(er))
                else:
                    feedback(0, okay_msg)
                finally:
                    if after: after()
                    
        # overwrite the run method             
        self.run = func_call
            
    def run(self):
        raise NotImplementedError('run')    
    
    def connect_button(self,        
        button: QWidget
      ) -> None:
        """ Connect the action to a button """
        button.clicked.connect(self.run)
        self._disconnection.append( button.clicked.disconnect )
    
    def connect_action(self,        
        action: QAction
      ) -> None:     
        """ Connect the action to a menu action """
        action.triggered.connect(self.run)
        self._disconnection.append( action.triggered.disconnect )
    
    def connect_checkbox(self, 
        checkbox
    ) -> None:
        """ Connect the action to a CheckBox """
        checkbox.stateChanged.connect(self.run)
        self._disconnection.append( checkbox.stateChanged.disconnect ) 
    
    def connect_combo(self, 
       combo: QComboBox, 
    ) -> None:
        """ Connect the action to a combo box """
        def combo_call(idx):
            self.run()
            
        combo.currentIndexChanged.connect(combo_call)
        self._disconnection.append( combo.currentIndexChanged.disconnect ) 
        
    def connect_item(self,         
        combo: QComboBox, 
        item_id: int        
     ) -> None:
        """ Connect the action to a combo box and the item id """
        try:
            f, item2action = self._combos[combo]
        except KeyError:
            self._new_combo(combo)
            f, item2action = self._combos[combo]        
        item2action[item_id] = self
        
    def _new_combo(self, combo):
        if combo in self._combos:
            try:
                combo.currentIndexChanged.disconnect()
            except TypeError:
                pass
                        
        item2action = {}
        def combo_call(i):
            try:
                action = item2action[i]
            except KeyError:
                return 
            action.run()            
        
        self._combos[combo] = (combo_call, item2action)
        combo.currentIndexChanged.connect(combo_call)
        self._disconnection.append(combo.currentIndexChanged.disconnect)

    
class Actions:
    def __init__(self):        
        self._disconnection = []
        self._combos = {}
        
    def add(self, func, inputs=None, feedback=None, before=None, after=None):
        return Action(self, func, inputs, feedback=feedback, before=before, after=after)
        
    def disconnect(self):
        for f in self._disconnection:
            try:
                f()
            except TypeError:
                pass
    

class WidgetControl:
    """ Allows to control a device widget 
    
    A WidgetControl is returned by the method .connect of the linker (:class:`BaseUiLinker`) 
    Is not realy intended to be used outiside of this scope. 
    
    it allows to:
       - update : update a widget with new data if this is not done automatically  
       - disable : the widget is disabled and also the corresponding nodes are free from the downloader
       - enable : re-enable the widget connection 
       - kill : clear any connections between widget and downloader and device so the widget can be destroyed 
    
    enable/disable can be use in the context of a hidden tab for instance.
    
    Args:
    
      Basically they are the linker and all arguments necessary to re-establisher a connection between
      the widget and Downloader and Device
      
        - linker :class:`BaseUiLinker` 
        - downloader  :class:`Downloader`the downloader used by linker to connect widget and device output
        - obj :class:`_BaseObject` device/interface/manager.. used by the linker to connect widget and action methods    
        - data Optional[BaseData]: As created by the linker connection 
   
    Examples:
    
    ::
    
        from pydevmgr_elt import open_device, Downloader
        from pydevmgr_qt import get_widget_factory
        
        downloader = Downloader()
        motor = open_device('resource_path.yml', 'motor1')
        w = get_widget_factory('ctrl', 'Motor').connect(downloader, motor)
        w.disable()
        w.enable()
        
    """
    def __init__(self, 
          linker, 
          downloader: Downloader, 
          device: _BaseObject, 
          data: BaseData 
        ) -> None:
        self.linker  = linker
        self.downloader = downloader
        self.device = device
        
        self.data   = data
         
    @property
    def widget(self):
        return self.linker.widget
    
    def update(self):
        """ Force the widget update now
        
        If self.config.link_update is false the update is not triggered by the downloader. One may want to 
        have more control on when the widget is updated
        """
        self.linker.update(self.data)
    
    def disable(self):
        """ disable the widget and disconnect it to the downloader 
        
        This means that all parameters used inside the widget are freed from the pool of node to 
        download. 
        The connection can be re-established using the ``enable`` method.
        
        disable/enable can be usefull when part of a widget is not used or hidden (for instance in a tab)
        """
        self.linker.disconnect()
        self.linker.widget.setEnabled(False)
        self.linker.widget.repaint()
    
    def enable(self):
        """ re-establish a connection previously cut by disable and enable the widget """
        self.linker.disconnect()
        self.linker.connect(self.downloader, self.device, self.data)
        self.linker.widget.setEnabled(True)
        
    def kill(self):
        """ Kill any connection between widget and downloader, so the widget can be destroyed """
        self.linker.disconnect()
        self.widget.setParent(None)
        self.linker = None
        self.device = None
        self.downloader = None
        self.data = None
    
    def append_in_layout(self, 
           layout: QLayout, 
           stretch: Optional[int] = 0, 
           alignment: Optional[int] = 0, 
           row: Optional[int] = 0, 
           column: Optional[int] = 0, 
           rowSpan: Optional[int] = 1, 
           columnSpan: Optional[int] = 1
        ) -> None:
        """ Place the attached ui widget in control into a QLayout """
        widget = self.widget 
        if isinstance(layout, QBoxLayout): 
            layout.addWidget(widget, stretch, QtCore.Qt.AlignmentFlag(alignment))
        elif isinstance(layout, QGridLayout):
            layout.addWidget(widget, row, column, rowSpan, columnSpan)
        else:
            layout.addWidget(widget)        
    
class BaseUiLinker:
    """ UiLinker takes care of the link between a widget and a downloader and device 
    
    Args:
       widget (optional, :class:`QWidget`) : 
               if not given the default widget is built with the Widget attribute 
                e.i.  MotorCtrl() is equivalent to MotorCtrl(MotorCtrl.Widget()) 
    
    Example:
    
    ::
       
       from pydevmgr_elt import open_device, Downloader
       from pydevmgr_qt import get_widget_factory
       
       downloader = Downloader()
       motor = open_device('resource_path.yml', 'motor1')
       
       motorCtrl = MotorCtrl().connect(downloader, motor)
           
    """        
    Data = BaseModel
    Widget = QWidget    
    class Config(BaseModel):
        link_failure: bool = True
        link_update: bool = True
    
    ## #############################################################
    #  
    #   Engine Methods
    #
    ## #############################################################
    def __init__(self, widget=None, config=None, **kwargs):

        self._config = self.parse_config(config, **kwargs)

        if widget is None:
            widget = self.new_widget(self._config)
        
        self.widget = widget
                
        self.actions = Actions()
        self.inputs = Inputs()
        self.outputs = Outputs()
        self.io = InputOutputs()        
        self.init_vars()        
    
    @property
    def config(self):
        return self._config
    
    @classmethod
    def new_widget(self, config):
        """ Return a new widget for this linker """
        return self.Widget()


    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            return cls.Config( **{**__config__, **kwargs} )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")

    def _disconnect_widget(self):
        # function to be overwrtiten by link
        pass
    
    def _link(self, 
            downloader_or_connection: Union[Downloader,Downloader.Connection],
            dl: DataLink, 
            data :  BaseData,
         ) -> None:
        """ Link the UI to a downloader 
        
        Each time the downloader will "download" the update will be executed with the 
        freshly updated data.
        
        Args:
            - downloader (:class:`pydevmgr.Downloader`)
            - dl (:class:`pydevmgr.DataLink`)
            - data (:class:`BaseModel`) (shall be the one linked with the dl DataLink)
        """
        self._disconnect_widget() # remove any connection 
        
        link_failure = self.config.link_failure
        link_update = self.config.link_update

        if isinstance(downloader_or_connection, Downloader):  
            connection  = downloader_or_connection.new_connection()
            new_connection = True 
        else:
            connection  = downloader_or_connection
            new_connection = False
        
        # The data is added to the queue of data to be downloaded 
        # y the downloader associated to the connection  
        #dl =  DataLink(obj, data)       
        connection.add_datalink(dl)
        
        if link_update:
            def download_callback(dummy=None):
                self.update(data)
            connection.add_callback(download_callback)
        
        if link_failure:
            connection.add_failure_callback(self.update_failure)
        
        ##
        #  Overwrite the disconnect_widget function 
        if new_connection:
            # we can simply disconnect the connection from the downloader all callback and datalink 
            # will be removed 
            def disconnect_widget():
                connection.disconnect()
        else:       
            # the connection may be used somewhere else so remove only the created 
            # datalinks and callback 
            if link_failure:
                def disconnect_widget():
                    connection.remove_datalink(dl)
                    if link_update:
                        connection.remove_callback(download_callback)
                    connection.remove_failure_callback(self.update_failure)
            else:
                def disconnect_widget():
                    connection.remove_datalink(dl)
                    if link_update:
                        connection.remove_callback(download_callback)
        self._disconnect_widget = disconnect_widget
        
    def connect(self, 
           downloader: Downloader, 
           obj: _BaseObject,
           data = None,       
        ) -> WidgetControl:
        """ Connect the widget to a obj 
        
        Args:
            downloader (:class:`pydevmgr_core.Downloader` or :class:`pydevmgr_core.DownloaderConnection`)
            obj  (:class:`pydevmgr_core._BaseObject`): most probably a  :class:`pydevmgr_core.BaseDevice`
            data (optional, :class:`pydantic.BaseModel`): data structure to be linked. If not given
                  It is created from the class .Data attribute
          
        Returns:
            ctrl: (:class:`WidgetControl`) This object allows to control the link between the widget, 
                the downloader and device. The ctrl is usefull for a parent app whish needs to 
                enable/disable widgets.
        """
        
        if data is None: 
            data = self.new_data()
        # self.disconnect() # Not sure about this, better to keep the disconnection manual 
        #self.setup_ui(obj, data)
        dl = DataLink(obj, data)
        self.setup_ui(obj, data)
        self._link(downloader, dl, data)          
        return WidgetControl(self, downloader, obj, data)
        
    def disconnect(self) -> None:
        """ Free the uilink from downloader connection callbacks 
        After executing, the refresh is stopped and widget actions are removed 
        """
        self._disconnect_widget()        
        self.disconnect_events()
    
    def new_data(self, **kwargs):
        return self.Data(**kwargs)
    
    def disconnect_events(self):
        # disconnect all button or other action associiated to a device 
        # which has been connected by setup              
        self.actions.disconnect()
    
    def update_failure(self, er):
        if er:
            self.widget.setEnabled(False) 
        else:
            self.widget.setEnabled(True)       
        
    def __del__(self):
        try:
            self.disconnect()
        except Exception as e:
            pass        

    ## #############################################################
    #
    #    Methods To be implemented 
    #
    ## #############################################################
    
    def init_vars(self):
        """ init all widget input and output handlers """
        pass
    
    def setup_ui(self, obj: _BaseObject, data: BaseData) -> None:
        """ setup the UI for a input device and associated data 
        
        setup change the static data and the actions associated to buttons, dropdown, etc ...
        """        
        # Disconnect all buttons, dropdoown, etc from previous connected events
        self.disconnect_events()
        # to be implemented 
        
        
    def update(self, data: BaseData) -> None:
        """ update the ui to new data 
        
        Args:
           data (class:`pydantic.BaseModel`): Data Model as returned by .new_data() method            
        """
        pass

    def feedback(self, er, msg=''):
        pass
 



class LayoutLinker:
    _wctrl_loockup = None
    Layout = QVBoxLayout
    _downloader = None
    def __init__(self, layout: Optional[str] = None):
        if layout is None:
            layout = self.Layout()
        
        self.layout = layout
        self._wctrl_loockup = {}
        
    def connect(self, downloader):
        """ Connect the LayoutLinker to a downloader """
        self._downloader = downloader
    
    def disconnect(self):
        """ disconnect from downloader and disconnect all linker children """
        self._downloader = None
        for l in self._wctrl_loockup.values():
            l.disable()
        
    def clear(self):
        """ Clear all widget and disconnect them """
        for i in reversed(range(self.layout.count())): 
            w = self.layout.itemAt(i).widget()
            w.setParent(None)            
            if w in self._wctrl_loockup:
                l = self._wctrl_loockup.pop(w)
                l.disable()
                
    def add_device(self, 
            device : _BaseObject, 
            widget_kind: str , 
            data: BaseData =None, 
            default_factory: Optional['WidgetFactory'] = None, 
            stretch: Optional[int] = 0, 
            alignment: Optional[int] = 0, 
            row: Optional[int] = 0, 
            column: Optional[int] = 0, 
            rowSpan: Optional[int] = 1, 
            columnSpan: Optional[int] = 1
            ):
        if self._downloader is None:
            raise ValueError("LayoutLinker is not connected")
        
        factory = get_widget_factory(widget_kind, device.config.type, default=default_factory)      
        linker = factory.build()
        ctrl = linker.connect(self._downloader, device, data=data)
        self._wctrl_loockup[ctrl.widget] = ctrl
        ctrl.append_in_layout(self.layout, stretch = stretch, alignment=alignment, row=row, 
                              column=column, rowSpan=rowSpan, columnSpan=columnSpan
                            )
                        
        
    
class WidgetFactory:
    """ A factory for widget and widget linker 
    
    A WdigetFactory is reutrned by :func:`get_widget_factory`
    
    Args:
        - widget_type (str) :  e.g. 'line', 'ctrl', 'cfg'
        - dev_type (str) : device type, e.g. 'Motor', 'Lamp', 'Adc', ...
        - LinkerClass (BaseUiLinker): Class of the linker. The linker is building the relation between
                                    a widget and the device. 
        - WidgetClass (QWidget, optional): if not given take the default defined in LinkerClass
        - DataClass (BaseModel, optional): the class to build the data structure. If not given take 
                                           one defined in LinkerClass 
    """
    def __init__(self, 
            widget_type: str, 
            dev_type: str, 
            LinkerClass: BaseUiLinker, 
            WidgetClass: Optional[QWidget] =None, 
            DataClass: Optional[BaseData] =None
          )-> None:
        self._widget_type = widget_type
        self._dev_type = dev_type 
        
        self.LinkerClass = LinkerClass
        if WidgetClass is None:
            WidgetClass = LinkerClass.Widget
        self.WidgetClass = WidgetClass
        if DataClass is None:
            self.DataClass = LinkerClass.Data
                
    def build_and_connect(self, 
          downloader: Downloader, 
          device: _BaseObject, 
          data: BaseData = None, 
          widget: Optional[QWidget] =None
        )-> WidgetControl:
        if widget is None:
            widget = self.WidgetClass()
        if data is None:
            data = self.DataClass()
            
        return self.Linker(widget).connect(downloader, device, data=data)
        
    def build(self,  widget: Optional[QWidget] =None) -> BaseUiLinker:
        if widget is None:
            widget = self.WidgetClass()
        return self.LinkerClass(widget)


_widget_factories = {}
def record_widget_factory(
       widget_type: str, 
       dev_type: Union[None, str,Iterable],
       LinkerClass: BaseUiLinker
     ) -> None:
    """ Record a widget linker defined by its type name and a device type 
    
    Args:
       - widget_type : string defining the widget (e.g. 'line', 'ctrl', 'cfg', ...)
       - dev_type : A list or a string for which the Linker apply. If None the Linker can handle any type 
       - LinkerClass: The Linker Class (derived from :class:`BaseUiLinker`)
    """
    
    global _widget_factories
    if not isinstance(dev_type, str) and hasattr(dev_type, '__iter__'):
        for dt in dev_type:    
            _widget_factories[(widget_type, dt)] =  WidgetFactory(widget_type, dt, LinkerClass)
    else:        
        _widget_factories[(widget_type, dev_type)] =  WidgetFactory(widget_type, dev_type, LinkerClass)

def get_widget_types(dev_type: str):
    """ Return a list of available widget str types for a given device name """
    return [wt for wt, dt in _widget_factories if dt==dev_type]
    
def get_widget_factory(
       widget_type: str, 
       dev_type: Union[None, str], 
       default: Optional[WidgetFactory] = None, 
       alt_dev_type: Optional[Iterable] = None
     ) -> WidgetFactory:
    """ Return a widget linker factory to build a widget for a given widget type and device type 
    
    Args:
    
       - widget_type (str) : string defining the widget (e.g. 'line', 'ctrl', 'cfg', ...)
       - dev_type (str): The device type 
       - default (optional, WidgetFactory): If given and the Factory is not found it is returned else
                an error is raised.
        - alt_dev_type (iterable, optional): Alternative type list if the widget is not found with dev_type 
    Example:
        
        > get_widget_factory('ctr', 'Motor').build().connect(downloader, motor1)
        
    """ 
    # it may be that this is defined for all device type
    try:
        return  _widget_factories[(widget_type, None)]
    except KeyError:
        pass

    try:
        return  _widget_factories[(widget_type, dev_type)]
    except KeyError:
        pass

    if alt_dev_type:
        for tpe in alt_dev_type:
            try:
                return  _widget_factories[(widget_type, tpe)]
            except KeyError:
                pass

    
    if default:
        return default

    raise ValueError(f"Widget Factory with type {widget_type} and device type {dev_type} cannot be found, set `alt_dev_type` or `default`")




from math import isinf
from pydevmgr_core._core_objects._class_recorder import get_class, KINDS
from pydevmgr_core._core_objects._core_base import _BaseObject, CONFIG_MODE, CONFIG_MODE_DEFAULT, _BaseObjDict, _BaseObjDictProperty
from pydevmgr_core._core_objects._core_pydantic import _default_walk_set

from typing import Any, Type, Tuple, Optional
#from ._class_recorder import get_class, KINDS
#from ._core_base import _BaseObject



from pydantic import BaseModel, validator
import weakref


class ObjDictConfig(BaseModel):
    kinds: set  = set()
    types: set  = set()
    exclude_kinds: set = set()
    exclude_types: set = set()
    default_kind: Optional[KINDS] = None
    default_type: Optional[str] = None

class ObjDictProperty(_BaseObjDictProperty):

    def __init__(self, name, cls=None, config=None, config_path=None, config_mode=CONFIG_MODE_DEFAULT):
        self._name = name
        self._cls = ObjDict if cls is None else cls
        if config is None:
            config = self._cls.Config()
        self._config = config 
        self._config_path = config_path
        self._config_mode = config_mode
    
    def get_config(self, parent):
        """ return configuration from parent object """
        # this has to be implemented for each kinds
        if self._config_path:
            config = getattr( parent.config, self._config_path )
        elif self._name:

            try:
                config = getattr(parent.config, self._name+"_rules")
            except AttributeError:
                config = self._config
            
             
    
        if config is not self._config:
            if self._config_mode == CONFIG_MODE.DEFAULT:
                _default_walk_set(self._config, config)
            else:
                for k,v in self._config.dict( exclude_unset=True ).items():
                    # TODO: Add warning for config overwrite 
                    setattr(config, k, v)
        return config   

    def __get__(self, parent, pcls=None):
        if parent is None: 
            return self

        try:
            return parent.__dict__[self._name]
        except KeyError:
            
            cfg_dict = getattr(parent.config, self._name, {})
                            
            output = self._cls( cfg_dict, __parent__= parent, __config__= self.get_config(parent)  )
            
            # Cash the dictionary in parent 
            parent.__dict__[self._name] = output 
            return output
    
    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name 


class ObjDict(_BaseObjDict):
    Config = ObjDictConfig

    def __init__(self, __d__ = {}, __parent__= None, __config__= None, **kwargs):
        
        if __config__ is None:
            config = self.Config()
        elif isinstance(__config__, dict):
            config = self.Config.parse_obj(__config__)
        else:
            config = __config__

        d = dict(__d__, **kwargs)
        super().__init__( {name:self.validate_child(name, child, config, __parent__) for name, child in d.items() }) 
        

        if __parent__:
            self._parent_ref = weakref.ref(__parent__) 
        else:
            self._parent_ref = lambda : None
        
        self._config = config
    def __setitem__(self, key, value):
        value = self.validate_child(key, value, self._config, self._parent_ref())
        super().__setitem__(key, value)
       
    def update(self, __d__={}, **kwargs):
        d = {name:self.validate_child(name, child, self._config, self._parent_ref()) for name, child in dict(__d__, **kwargs).items()}
        super().update( d )
    

    @classmethod
    def validate_kind(cls, name, kind, config):
        if kind is None:
            if config.default_kind:
                return config.default_kind
            else:
                raise ValueError(f"children {name!r} no default kind")
            
        if config.kinds:
            if kind is None:
                return config.kinds[0]
            if (kind not in config.kinds) or (kind in config.exclude_kinds):
                raise ValueError(f"children {name!r}  does not have the right kind")
        elif kind in config.exclude_kinds: 
            raise ValueError(f"children {name!r}  does not have the right kind")
        return kind
    
    @classmethod
    def validate_type(cls, name, type_, config):
        if type_ is None:
            if config.default_type:
                return config.default_type
            else:
                raise ValueError(f"children {name!r} no default kind")
           
        if config.types:    
            if (type_ not in config.types) or (type_ in config.exclude_types):
                raise ValueError(f"children {name!r}  does not have the right type")
        elif type_ in config.exclude_types: 
            raise ValueError(f"children {name!r}  does not have the right type")
        return type_
     
    @classmethod
    def validate_config(cls, name, conf: BaseModel, config: BaseModel)-> Tuple[Type,BaseModel]:
        cls.validate_kind(name, getattr(conf, 'kind', None), config )
        cls.validate_type(name, getattr(conf, 'type', None), config )
        ObjClass = get_class(conf.kind, conf.type)
        return ObjClass, conf     
    
    @classmethod
    def validate_dict(cls, name, d: dict, config: BaseModel)-> Tuple[Type,BaseModel]:
        kind  = cls.validate_kind( name, d.get("kind", None), config )
        type_ = cls.validate_type( name, d.get("type", None), config )
        ObjClass = get_class(kind, type_)
        return ObjClass, ObjClass.Config.parse_obj(d)
 
    @classmethod
    def validate_object(cls, name, obj: Any, config):
        if not isinstance( obj, _BaseObject ):
            raise ValueError(f"child {name} of of type {type(obj)} is not allowed")
        cls.validate_kind( name, obj.config.kind, config )
        cls.validate_type( name, obj.config.type, config )

        return obj
    
  
    @classmethod
    def validate_child(cls, name, child, config, parent=None):
        if isinstance( child, _BaseObject):
            return cls.validate_object(name, child, config) 
        
        if isinstance(child, BaseModel):
            ObjClass, config = cls.validate_config( name, child, config )
        
        elif isinstance( child, dict ):
            ObjClass, config = cls.validate_dict( name, child, config )

        else:
            raise ValueError(f"Unexpected child type: {type(child)}")
        if parent:
            obj = ObjClass.new(parent, name, config=config)
        else:
            obj = ObjClass(key=name, config=config)
        return obj
    
    @classmethod
    def new(cls, parent, name=None, **kwargs):
        
        config = cls.Config(**kwargs)
         
        if name is None:
            cfg_dict = {}
        else:
            cfg_dict = getattr(parent.config, name)
        return cls( cfg_dict, __parent__= parent, __config__= config )


    @classmethod
    def prop(cls, name=None, **kwargs):
        config = cls.Config(**kwargs)
        return ObjDictProperty(name, cls, config=config)



if __name__ == "__main__":
    from pydevmgr_core import BaseDevice, BaseManager, record_class
    from pydevmgr_core._core_objects._core_pydantic import GenDevice
    from typing import Dict 

    @record_class
    class D(BaseDevice):
        class Config(BaseDevice.Config):
            type = "D"
            num: int = 0 
    
    class M(BaseManager):
        class Config(BaseManager.Config):
            devices: Dict[str, GenDevice] = { 'dev1': D.Config(num=1)}

        devices = ObjDict.prop('devices') 

    assert M().devices['dev1'].config.num == 1



from ._core_node import BaseNode, NodesReader, NodesWriter
from ._core_base import kjoin, _BaseObject, new_key, path 
from typing import Union, List, Optional, Any, Dict, Callable
from pydantic import create_model, validator
from inspect import signature , _empty


class NodeAliasConfig(BaseNode.Config):
    type: str = "Alias"
    nodes: Optional[Union[List[Union[str, tuple]], str]] = None
    
 
class NodeAlias1Config(BaseNode.Config):
    type: str = "Alias1"
    node: Optional[Union[str,tuple]] = None
    


class NodeAliasProperty(BaseNode.Property):
    # redefine the node alias property to explicitly add the nodes argument
    def __init__(self, cls, constructor, name, nodes, *args, **kwargs):
        super().__init__( cls, constructor, name, *args, **kwargs)
        self._nodes = nodes
    
    def new(self, parent):
        config = self.get_config(parent)
        if self._name is None:
            name = new_key(config)
        else:
            name = self._name    
        obj = self._constructor(parent, name, self._nodes, *self._args, config=config, **self._kwargs)            
        self._finalise(parent, obj)
        return name, obj 
        
            
class NodeAlias(BaseNode):
    """ NodeAlias mimic a real client Node. 
        
    The NodeAlias object does a little bit of computation to return a value with its `get()` method and 
    thanks to required input nodes.
     
    The NodeAlias cannot be use as such without implementing a `fget` method. This can be done by 
    implementing the fget method on an inerated class or with the `nodealias` decorator. 
    
    NodeAlias is an abstraction layer, it does not do anything complex but allows uniformity among ways to retrieve values. 
    
    NodeAlias object can be easely created with the @nodealias() decorator
    
    Args:
        key (str): Key of the node
        nodes (list, class:`BaseNode`): list of nodes necessary for the alias node. When the 
                     node alias is used in a :class:`pydevmgr_core.Downloader` object, the Downloader will automaticaly fetch 
                     those required nodes from server (or other node aliases).
                     
    Example: 
    
    ::
    
        >>> is_inpos_for_test = NodeAlias('is_inpos_for_test', [mgr.motor1.stat.pos_actual])
        >>> is_inpos_for_test.fget = lambda pos: abs(pos-4.56)<0.01
        >>> is_inpos_for_test.get()
    
    :: 
    
        @nodealias("is_all_standstill", [mgr.motor1.stat.substate, mgr.motor2.stat.substate])
        def is_all_standstill(m1_substate, m2_substate):
            return m1_substate == Motor.SUBSTATE.OP_STANDSTILL and m2_substate == Motor.SUBSTATE.OP_STANDSTILL
    
        >>> is_all_standstill.get()
        True
        
        >>> downloader = Downloader( [is_all_standstill] )
        >>> downloader.download()        
        >>> downloader.get_data()
        {'fcs.motor1.substate': 100,
         'fcs.motor2.substate': 100,
         'is_all_standstill': False}
         
    In the exemple above one can see that the mgr.motor/1/2.stat.substate has been automatically added 
    to the nodes to be fetched from OPC-UA server(s). 
    
    Here is an exemple of customized NodeAlias, it return the mean and the max of a value, updated
    at each get :
    
    ::
    
        import numpy as np 
        
        class MinMaxNode(NodeAlias):
            min = +np.inf
            max = -np.inf
            
            def fget(self, pos):
                self.min = min(pos, self.min)
                self.max = max(pos, self.max)
                return ( self.min , self.max )
            
            def reset(self):
                self.min = +np.inf
                self.max = -np.inf
                
        mot1_minmax = MinMaxNode( "minmax",  [mgr.motor1.stat.pos_actual])
                
    .. seealso::  
        :func:`nodealias`
        :func:`nodealiasproperty`
        :class:`NodeAlias`            
    """
    Config = NodeAliasConfig
    Property = NodeAliasProperty
    
    _n_nodes_required = None
    _nodes_is_scalar = False
    def __init__(self, 
          key: Optional[str] = None, 
          nodes: Union[List[BaseNode], BaseNode] = None,
          config: Optional[Config] = None,
          **kwargs
         ):
         
        super().__init__(key, config=config, **kwargs)
        if nodes is None:
            nodes = []
        
        elif isinstance(nodes, BaseNode):
            nodes = [nodes]
            self._nodes_is_scalar = True
        if self._n_nodes_required is not None:
            if len(nodes)!=self._n_nodes_required:
                raise ValueError(f"{type(self)} needs {self._n_nodes_required} got {len(nodes)}")
        self._nodes = nodes
    
    @property
    def sid(self):
        """ sid of aliases must return None """ 
        return None
    
    @property
    def nodes(self):
        return self._nodes
    
    @classmethod
    def prop(cls, name: Optional[str] = None, nodes=None, **kwargs):
        nodes = [] if nodes is None else nodes
        #config = cls.Config.parse_obj(kwargs)  
        config = cls.parse_config(kwargs)
        return cls.Property(cls, cls.new, name, nodes, config=config)
                
    @classmethod
    def new(cls, parent, name, nodes=None, config=None, **kwargs):
        """ a base constructor for a NodeAlias within a parent context  
        
        The requirement for the parent :
            - a .key attribute 
            - attribute of the given name in the list shall return a node
        """
        config = cls.parse_config(config, **kwargs)
        if nodes is None:
            if config.nodes is None:
                nodes = []
            else:
                nodes = config.nodes
        # nodes = config.nodes                
        # handle the nodes now
        #if nodes is None:
        #    raise ValueError("The Node alias does not have origin node defined, e.i. config.nodes = None")
        if isinstance(nodes, str):
            nodes = [nodes]
        elif hasattr(nodes, "__call__"):
            nodes = nodes(parent)
                                
        parsed_nodes, node_names = zip(*(cls._parse_node(parent, n) for n in path(nodes)))
        
        return cls(kjoin(parent.key, name), parsed_nodes, config=config, localdata=parent.localdata, node_names=node_names)
    
    @classmethod
    def _parse_node(cls, parent: _BaseObject, in_node: Union[tuple,str,BaseNode]) -> 'NodeAlias':
        if isinstance(in_node, BaseNode):
            return in_node, in_node.key
        
        
        if isinstance(in_node, str):
            in_node, _, surname = in_node.partition(" as ")
            
            node = getattr(parent, in_node)
            if not isinstance(node, BaseNode):
                raise ValueError("Attribute {!r} of parent is not node got a {}".format(in_node, type(node)))
            return node, surname or in_node        
        
        if isinstance(in_node, tuple):
            cparent = parent
            for sn in in_node[:-1]:
                cparent = getattr(cparent, sn)
            
            name, _, surname = in_node[-1].partition(" as ")
            node = getattr(cparent, name)
            if not isinstance(node, BaseNode):
                raise ValueError("Attribute {!r} of parent is not a node got a {}".format(in_node , type(node)))
            return node, name or surname
            
        raise ValueError('node shall be a parent attribute name, a tuple or a BaseNode got a {}'.format(type(in_node)))         
        
    def get(self, data: Optional[Dict] =None) -> Any:
        """ get the node alias value from server or from data dictionary if given """
        if data is None:
            _n_data = {}
            NodesReader(self._nodes).read(_n_data)
            values = [_n_data[n] for n in self._nodes]
            #values = [n.get() for n in self._nodes]
        else:
            values = [data[n] for n in self._nodes]
        return self.fget(*values)
    
    def set(self, value: Any, data: Optional[Dict] =None) -> None:
        """ set the node alias value to server or to data dictionary if given """
        values = self.fset(value)
        if data is None:
            NodesWriter(dict(zip(self._nodes, values))).write()                        
            #for n,v in zip(self._nodes, values):
            #    n.set(v)
        else:
            for n,v in zip(self._nodes, values):
                data[n] = v        
    
    def fget(self, *args) -> Any:
        """ Process all input value (taken from Nodes) and return a computed value """
        raise NotImplementedError('fget')
    
    def fset(self, value) -> Any:
        """ Process one argument and return new values for the aliased Nodes """
        raise NotImplementedError('fset')    




class NodeAlias1(BaseNode):
    Config = NodeAlias1Config
    Property = NodeAliasProperty
    
    def __init__(self, 
          key: Optional[str] = None, 
          node: Optional[BaseNode] = None,
          config: Optional[Config] = None, 
          localdata: Optional[dict] = None, 
          node_name: Optional[str] = None, 
          **kwargs
         ):        
        super().__init__(key, config=config, localdata=localdata, **kwargs)    
        if node is None:            
            raise ValueError("the node pointer is empty, alias node cannot work without")    
                    
        self._node = node
        self._node_name = node_name
    
    @property
    def sid(self):
        """ sid of aliases must return None """ 
        return None
    
    @property
    def node(self):
        return self._node
    
    # nodes property is mendatory for the NodeReader 
    @property
    def nodes(self):
        return [self._node]
    
    @classmethod
    def prop(cls, 
          name: Optional[str] = None, 
          node: Union[BaseNode,str] = None,  
          **kwargs
        ) -> NodeAliasProperty:
        # config = cls.Config.parse_obj(kwargs)
        config = cls.parse_config(kwargs)
        return cls.Property(cls, cls.new, name, node, config=config)
    
    @classmethod
    def new(cls, parent, name, node=None,  config=None, **kwargs):
        """ a base constructor for a NodeAlias within a parent context  
        
        The requirement for the parent :
            - a .key attribute 
            - attribute of the given name in the list shall return a node
        """
        config = cls.parse_config(config, **kwargs)
        if node is None:            
            node = config.node 
        if node is None:
            raise ValueError("node origin pointer is not defined")                             
        parsed_node, node_name = NodeAlias._parse_node(parent, path(node))    
        
        return cls(kjoin(parent.key, name), parsed_node, config=config, localdata=parent.localdata, node_name=node_name)
    
    
    def get(self, data: Optional[Dict] =None) -> Any:
        """ get the node alias value from server or from data dictionary if given """
        if data is None:
            _n_data = {}
            NodesReader([self._node]).read(_n_data)
            value = _n_data[self._node]
        else:
            value = data[self._node]
        return self.fget(value)    
        
    def set(self, value: Any, data: Optional[Dict] =None) -> None:
        """ set the node alias value to server or to data dictionary if given """
        value = self.fset(value)
        if data is None:
            NodesWriter({self._node:value}).write()                        
            #for n,v in zip(self._nodes, values):
            #    n.set(v)
        else:
            data[self._node] = value
    
    def fget(self,value) -> Any:
        """ Process the input retrieved value and return a new computed on """
        return value
    
    def fset(self, value) -> Any:
        """ Process the value intended to be set """
        return value


def nodealiasproperty(name, nodes, *args, **kwargs):
    """ A decorator for a quick alias node creation 
    
    This shall be implemented in a parent interface or any class with the ``get_node`` method
    
    Args:
        name (str) : name of the node. The key of the node will be parent_key.name
        nodes (iterable): List of a mix of node or string coresponding to the parent attribute pointing
                          to a node.  
        \*args, \**kwargs: All other arguments necessary for the node construction this is only used 
                         if an laternative cls is given 
    """
    return NodeAlias.prop(name, nodes, *args, **kwargs)


def nodealias1property(name, node, *args, **kwargs):
    """ A decorator for a quick alias node creation 
    
    This shall be implemented in a parent interface or any class with the ``get_node`` method
    
    Args:
        name (str) : name of the node. The key of the node will be parent_key.name
        node (:class:`BaseNode`, str): Either a node or a string coresponding to the parent attribute 
                pointing to a node.  
        \*args, \**kwargs: All other arguments necessary for the node construction 
    """
    return NodeAlias1.prop(name, node, *args, **kwargs)



def nodealias(key: Optional[str] =None, nodes: Optional[list] = None):
    """ This is a node alias decorator 
    
    This allow to quickly embed any function in a node without having to subclass the Alias Node
    
    The build node will be readonly, for a more complex behavior please subclass a NodeAlias
    
    Args:
        key (str): string key of the node
        nodes (lst): list of nodes 
        
    Returns:
       func_setter: a decorator for the fget method  
       
    Example:
    
    A simulator of value:
    
    ::
        
        # To be replaced by real stuff of course
        @node('temperature')
        def temperature():
            return np.random.random()*3 + 20
        @node('motor_pos')
        def motor_pos():
            return np.random.random()*100
        
        # the nodealias focus is computed from temperature and motor position 
        @nodealias('focus', [temperature, motor_pos]):
        def focus(temp, pos):
            return pos+ 0.45*temp + 23.
            
    In the example above when doing `focus.get()` it will automaticaly fetch the `temperature` and
    `motor_pos` nodes.  
        
    """
    node = NodeAlias(key,nodes)
    def set_fget(func):
        node.fget = func
        if hasattr(func, "__func__"):
            node.__doc__= func.__func__.__doc__
        else:
            node.__doc__ = func.__doc__
        return node
    return set_fget



def nodealias1(key: Optional[str] = None, node: Optional[BaseNode] = None) -> Callable:
    """ This is a node alias decorator 
    
    This allow to quickly embed any function in a node without having to subclass the Alias Node
    This is the counterpart of :func:`nodealias` except that it explicitely accept only one node
    as input instead of severals
    
    The build node will be readonly, for a more complex behavior please subclass a NodeAlias
    
    Args:
        key (str, optional): string key of the node
        node (:class:`BaseNode`): input node. This is not optional and shall be defined 
                                  It is however after key for historical reason 
        
    Returns:
       func_setter: a decorator for the fget method  
       
    Example:
    
    A simulator of value:
    
    ::
        
        @node('temperature_volt')
        def temperature_volt():
            return np.random.random()*3 + 20
        
        @nodealias1('temperature_celcius', temperature_volt):
        def temperature_celcius(temp_volt):
            return temp_volt*12.3 + 2.3
            
    """
    node = NodeAlias1(key,node)
    def set_fget(func):
        node.fget = func
        if hasattr(func, "__func__"):
            node.__doc__= func.__func__.__doc__
        else:
            node.__doc__ = func.__doc__
        return node
    return set_fget

def to_nodealias_class(_func_: Callable =None, *, type: Optional[str] = None):
    if _func_ is None:
        def node_func_decorator(func):
            return _nodealias_func(func, type)
        return node_func_decorator
    else:
        return _nodealias_func(_func_, type)
    
    
def _nodealias_func(func, type_):
    if not hasattr(func, "__call__"):
        raise ValueError(f"{func} is not callable")
    
    try:    
        s = signature(func)
    except ValueError: # assume it is a builtin class with one argument 
        conf_args = {}
        obj_args = []
        
    else:
        
        conf_args = {}
        obj_args = []
        poasarg = 0
        for a,p in s.parameters.items():
            if p.default == _empty:
                if a in ['com', 'localdata', 'key', 'nodes']:
                    if poasarg:
                        raise ValueError("Pos arguments must be after or one of 'com', 'localdata' or 'key'")
                    obj_args.append(a)
                else:
                    poasarg += 1        
            else:                
                if p.annotation == _empty:
                    conf_args[a] = p.default 
                else:
                    conf_args[a] = (p.annotation,p.default)
                    
    extras = {}
    if type_ is None:        
        if  "type" in conf_args:
            type_  = conf_args['type']            
        else:
            type_ = func.__name__
            extras['type'] = type_
    else:
        extras['type'] = type_
        
          
                
    Config = create_model(type_+"Config", **extras, **conf_args, __base__=NodeAlias.Config)
        
    if conf_args or obj_args: 
        conf_args_set = set(conf_args)        
                
        if obj_args:        
            def fget_method(self, *args):
                c = self.config
                return func(*[getattr(self,__a__) for __a__ in obj_args], *args, **{a:getattr(c,a) for a in conf_args_set})
        else:
            def fget_method(self, *args): 
                c = self.config           
                return func(*args, **{a:getattr(c,a) for a in conf_args_set})        
            
    else:
        def fget_method(self, *args):
            return func(*args)
    try:
        doc = func.__doc__
    except AttributeError:
        doc = None
                   
    return type(type_+"NodeAlias", 
                (NodeAlias,), 
                {'Config':Config, 'fget': fget_method, '__doc__':doc, '_n_nodes_required': poasarg}
               )
    



    




from typing import Any, Tuple, Optional, List,  Union, Type, TypeVar, Iterator
from pydantic import BaseModel, Extra,  validator, create_model, root_validator, ValidationError
from pydantic.class_validators import root_validator

from ._class_recorder import get_class, KINDS
from ._core_model_var import StaticVar, NodeVar

from ._core_pydantic import _default_walk_set
from enum import Enum 
import yaml
from ..io import ioconfig, load_config, parse_file_name
import logging 
import weakref

log = logging.getLogger('pydevmgr')

class CONFIG_MODE(str, Enum):
    OVERWRITE = "OVERWRITE"
    DEFAULT = "DEFAULT"


AUTO_BUILD_DEFAULT = True
CONFIG_MODE_DEFAULT = CONFIG_MODE.DEFAULT


ObjVar = TypeVar('ObjVar')

    
class BaseConfig(BaseModel):
    kind: KINDS = ""
    type: str = ""

    
    class Config: # this is the Config of BaseModel
        extra = Extra.forbid
        validate_assignment = True   
        use_enum_values = True 
    

    
    @root_validator(pre=True)
    def _base_root_validator(cls, values):
        # check is there is a cfgfile defined. if defined load the corresponding file and feeds its 
        # content inside values (unless already defined in value)

        # This is important to pop the cfgfile, so this will not be executed every time there is 
        # an attribute assignment 
        cfgfile = values.pop('cfgfile', None) 

        if cfgfile:
            cfg = load_config(cfgfile)
            for k, v in cfg.items():
                values.setdefault(k,v)
        return values

    def _parent_class_ref(cls):
        return None
    

    def _get_parent_class(self):
        p = self._parent_class_ref()
        if p is not None:
            return p
        return get_class(self.kind, self.type)

    @validator('kind')
    def _kind_validator(cls, kind):
        return cls.validate_kind(kind)


    @classmethod
    def validate_kind(cls, kind):
        if kind:
            return KINDS(kind)
    
    @classmethod
    def validate_type(cls, type_):
        return type_
    

    @classmethod
    def from_cfgfile(cls, cfgfile, path: str = ''):
        """ Create the config object from a yaml config file 
        
        Args:
            cfgfile (str): Configuration file relative to one of $CFGPATH or absolute path 
            path (str, optional): path where to find the root configuration. if "" the config file 
                 contains the root. For instance 'a.b.c' will look at configuration in cfg['a']['b']['c']
        """
        config = load_config(cfgfile)
        if path is not None:
            config = path_walk_item(config, path)
        return cls.parse_obj(config)                    
    
    

class ChildrenCapabilityConfig(BaseModel): 
    auto_build: bool = AUTO_BUILD_DEFAULT 

    @root_validator(pre=False)
    def _root_post_validator(cls, values):
        """ Every dict with kind and type member will be transformed to its right class 

        This will only be used if extra="allow"
        """
        errors = []
        for k,v in values.items():
            if k in cls.__fields__: # field exist in the class and will be treated independently 
                continue
            try:
                values[k] = cls.validate_extra(k, v, values)
            except ValueError as e:
                errors.append(e)
        if errors:
            raise ValueError( "\n\n  ".join(str(e) for e in errors))
        return values
    
    @classmethod
    def validate_extra(cls, name, extra, values):
        if isinstance(extra, dict) and "kind" in extra and "type" in extra:
            ObjClass = get_class(extra["kind"], extra["type"])
            return ObjClass.Config.parse_obj(extra)
        
        if getattr( cls.Config, "extra_obj_only", False):
            if isinstance( extra, _BaseObject.Config):
                return extra
            raise ValueError(f"""Extra {name!r} of type {type(extra)} not allowed""") 
        return extra


def path_walk_item(d, path):
    if isinstance(path, int):
        # get the first key
        
        if hasattr(d, "keys"):
            k = list(d.keys())[path]
            return d[k]
        else:
            return d[path]
   
    if path:
        for p in (s for s in  path.split('.') if s):
            d = d[p]
    return d


def path_walk_attr(obj, path):
    for p in (s for s in path.split('.') if s):
        obj = getattr(obj, p)
    return obj


def _path_name(d, path):
    if isinstance(path, int):
        if hasattr(d, "keys"):
            k = list(d.keys())[path]
            return k
        else:
            return None
    if isinstance(path, str):
        _, name = ksplit(path)
        return name 
    return None


def _get_class_dict( d, default_type: Optional[Union[str, Type]] = None):

    try:
        kind = d['kind']
    except KeyError:
        raise ValueError('"kind" attribute missing')
    
    
    try:
        st = d['type']
    except KeyError:
        raise ValueError('"type" attribute missing')    

    try:
        return get_class(kind, st)
    except ValueError as e:
        if default_type:
            if isinstance(default_type, type):
                return default_type
            return get_class(kind, default_type)
        raise e


def _get_class_config( c, default_type: Optional[Union[str, Type]] = None):
    try:
        return get_class(c.kind, c.type)
    except ValueError as e:
        if default_type:
            if isinstance(default_type, type):
                return default_type
            return get_class(c.kind, default_type)
        raise e
  




def open_class(
        cfgfile: str, 
        path: Optional[Union[str, int]] = None, 
        default_type: Optional[str] = None,
        **kwargs
        ):
    """ open a pydevmgr class and configuration object from a config file 

    Args:
        cfgfile: relative to on of the $CFPATH or absolute path to yaml config file 
        kind (optional, str): object kind as enumerated in KINDS ('Manager', 'Device', 'Interface', 'Node', 'Rpc')
            if None look inside the configuration file and raise error if not defined. 
        
        path (optional, str, int): an optional path to find the configuration inside the config file 
             'a.b.c' will go to cfg['a']['b']['c']
             If an integer N will get the Nth element of the cfgfile 

        default_type (optional, str): A default type if no type is defined in the configuration file
            If default_type is None and no type is found an error is raised 

    Returns:
        ObjClass :  An pydevmgr Object class (Manager, Device, Node, Rpc, Interface)
        config : An instance of the config (BaseModel object)
        pname (str, None) : The name of the object extracted from the path 
    """
    allconf = load_config(cfgfile)
    
    pname = _path_name(allconf, path)
    allconf = path_walk_item(allconf, path)
    
    allconf.update( kwargs )
    
    try:
        kind = allconf['kind']
    except KeyError:
        raise ValueError("Configuration has no 'kind' defined")

    tpe = allconf.get('type', default_type)
    if not tpe:
        raise ValueError(f"Cannot resolve {kind} type")
    Object = get_class(kind, tpe)
    config = Object.Config.parse_obj(allconf)
    return Object, config, pname 

def open_object(
        cfgfile,
        key: Optional[str]= None, 
        path: Optional[Union[str, int]] = None, 
        prefix: str = '', 
        default_type: Optional[str] = None, 
        **kwargs
    ):
    """ open an object from a configuration file 
    Args:
        cfgfile: relative to on of the $CFPATH or absolute path to yaml config file 
        kind (optional, str): object kind as enumerated in KINDS ('Manager', 'Device', 'Interface', 'Node', 'Rpc')
            if None look inside the configuration file and raise error if not defined. 
        
        path (optional, str): an optional path to find the configuration inside the config file 
             'a.b.c' will go to cfg['a']['b']['c']
        default_type (optional, str): A default type if no type is defined in the configuration file
            If default_type is None and no type is found an error is raised

    Returns:
        obj : instanciatedpydevmgr object (Manager, Device, Node, Rpc, Interface)

    """
    Object, config, pname = open_class(cfgfile, path=path, default_type=default_type, **kwargs)

    if key is None and pname:
        key = kjoin(prefix, pname)
    return Object(key, config=config)
        
class BaseData(BaseModel):
    # place holder for Data class 
    key: StaticVar[str] = ""


def load_yaml_config(yaml_payload: str, path: Optional[Union[str, tuple]] = None) -> Tuple[Type,BaseConfig]:
    """ Load a yaml configuration and pare it in the right configuration object 
    
    The Config class used is localised thanks to the `kind` and `type` string argument inside the yaml
    If the class is not recognised an Exception is raised.
    
    Args:
        yaml_payload (str):  The yaml string payload 
        path (optional, str, tuple):  A string or tuple representing a path through the wanted object 
    
    Returns:
        cls (Type):  The object Class 
        config (BaseModel):  The parsed configuration 
    """
    payload = yaml.load(yaml_payload, Loader=ioconfig.YamlLoader)
    
    payload = path_walk_item(payload, path)
   
    return load_dict_config(payload)
    
def load_dict_config(payload: dict)-> Tuple[Type,BaseConfig]:
    try:
        kind = payload['kind']
    except KeyError:
        raise ValueError('"kind" attribute missing')
    try:
        type = payload['type']
    except KeyError:
        raise ValueError('"type" attribute missing')    
    
    cls = _get_class_dict(payload)
    return cls, cls.parse_config(payload)

def build_yaml(yaml_payload, key: Optional[str]=None, *args, **kwargs):
    """ Build and return an object from its yaml configuration payload """
    cls, config = load_yaml_config(yaml_payload)
    return cls(key, *args, config=config, **kwargs)

def load_and_build( cfgfile: str, key: Optional[str] = None, *args, **kwargs):
    """ Load a config file and build the conrespoding object defined by kind and type config parameters """
    payload = load_config(cfgfile)
    
    cls, config = load_dict_config(payload)
    return cls(key, *args, config=config, **kwargs)
    
def kjoin(*args) -> str:
    """ join key elements """
    return ".".join(a for a in args if a)

def ksplit(key: str) -> Tuple[str,str]:
    """ ksplit(key) ->  prefix, name
    
    >>> ksplit('a.b.c')
    ('a.b', 'c')
    """
    s, _, p = key[::-1].partition(".")
    return p[::-1], s[::-1]

def path( keys: Union[str, list])-> Union[str, List[Union[str, tuple]]]:
    if isinstance(keys, str):
        l = keys.split(".")
        if len(l)<2:
            return keys
        return tuple(l)
    elif isinstance(keys, tuple):
        return keys
    elif hasattr( keys, '__iter__' ):
        return [path(k) for k in keys]
    return keys 
    #raise ValueError(f'expecting a string or an iterable on string got a {type(keys)}')

    


def reconfig(ConfigClass: Type, config: BaseConfig, kwargs: dict) -> BaseConfig:    
    if config is None:
        return ConfigClass.parse_obj(kwargs)
    if isinstance(config, dict):
        return ConfigClass.parse_obj(dict(config, **kwargs))    
    return config


_key_counter = {}
def new_key(config):
    k = config.type+config.kind
    c = _key_counter.setdefault(k,0)
    c += 1
    _key_counter[k] = c
    return f'{k}{c:03d}'


def auto_build( obj, attr ):

    try:
        c = getattr( obj.config, attr  )
    except AttributeError:
        raise AttributeError(f"{attr!r} attribute is not a valid pydevmgr object. Nothing in .config matching")
    if not isinstance(c, BaseConfig):
        raise AttributeError(f"found {attr!r} in config but does not seems to be a pydevmgr config object")
    NewClass = c._get_parent_class()
    new = NewClass.new( obj, attr, config=c )
    obj.__dict__[attr] = new
    return new 
    


class _BaseProperty:    
    """ A Property is basically calling a constructor with dynamical and static arguments 
    
    The Property has the followind signature : 
        Property(constructor, name, *args, config=None, **kwargs)
        
    The constructor is called with the following signature
        constructor( parent , name, *args, config=config, **kwargs)
    
    So it must have at least twho positional arguments a config keyword argument and 
    optional keyword arguments  
    
    """    
    def __init__(self, cls, constructor, name, *args, config=None, config_path=None, config_mode=CONFIG_MODE_DEFAULT, **kwargs):
        
        self._cls = cls 
        self._constructor = constructor
        self._name = name         
        
        self._config = config 
        self._config_path = config_path
        self._config_mode = config_mode 
        self._args = args
        self._kwargs = kwargs
        
    @property
    def congig(self):
        return self._config
            
    def _finalise(self, parent, obj):
        pass
    
    
    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name 


    def __get__(self, parent, clp=None):
        if parent is None:
            return self 
        # try to retrieve the node directly from the parent __dict__ where it should 
        # be cached. If no boj cached create one and save it/ 
        # if _name is the same than the attribute name in class, this should be called only ones
        if self._name is None:
            try:
                obj = parent.__dict__[self]
            except KeyError:
                name, obj = self.new(parent)     
                # store in parent with self
                parent.__dict__[self] = obj   
        else:
            try:
                obj = parent.__dict__[self._name]
            except KeyError:
                name, obj = self.new(parent)     
                # store in parent with the name 
                parent.__dict__[self._name] = obj   

        return obj
                
    def get_config(self, parent):
        """ return configuration from parent object """
        # this has to be implemented for each kinds
        if self._config_path:
            config = getattr( parent.config, self._config_path )
        elif self._name:
            try:
                config = getattr(parent.config, self._name)
            except AttributeError:
                config = self._config
        
        if config is not self._config:
            if not isinstance( config, type(self._config) ):
                log.warning( f"The configuration Class missmatch in property {self._name!r} " )

            if self._config_mode == CONFIG_MODE.DEFAULT:
                _default_walk_set(self._config, config)
            else:
                for k,v in self._config.dict( exclude_unset=True ).items():
                    log.warning( f"config var {k} overwriten by property" )
                    setattr(config, k, v)
        return config    
            
                
    def new(self, parent):     
        config = self.get_config(parent)
        if self._name is None:
            name = new_key(config) 
            #name = config.kind+str(id(config))
        else:
            name = self._name                            
        obj = self._constructor(parent, name, *self._args, config=config, **self._kwargs)            
        self._finalise(parent, obj)
        return name, obj 
        

class _BaseObject:
    __all_cashed__ = False
    Config = BaseConfig
    Property = _BaseProperty    
    _config = None
    

    def __init_subclass__(cls, **kwargs) -> None:
         # if kwargs:
        cls.Config = create_model(  cls.__name__+".Config",  __base__=cls.Config, **kwargs)
        cls.Config._parent_class_ref = weakref.ref(cls)
            


    def __init__(self, 
          key: Optional[str] = None,  
          config: Optional[Config] = None, *,          
          localdata: Optional[dict] = None, 
          **kwargs 
        ) -> None:
        self._config = self.parse_config(config, **kwargs)                            
        if key is None: 
            key = new_key(self._config)
        
        self._key = key     
        self._localdata = localdata
        
    def __repr__(self):
        return "<{} key={!r}>".format(self.__class__.__name__, self._key)
    
    
    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            d = {**__config__, **kwargs} 
            return cls.Config( **d )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")
            
    @classmethod
    def new_args(cls, parent, config: Config) -> dict:
        """ build a dictionary of dynamical variables inerited from a parent """
        return dict( localdata = getattr(parent, "localdata", None) )
            
    @classmethod
    def new(cls, parent, name, config=None, **kwargs):
        """ a base constructor for BaseNode 
        
        parent must have the .key attribute 
        """        
        # here shall be implemented something to deal with config, it might be that config it comming
        # from the parent.config 
        config = cls.parse_config(config, **kwargs)
        if name is None:
            name = new_key(config)                                
        return cls(kjoin(parent.key, name), config=config, **cls.new_args(parent, config))
    
    @classmethod
    def prop(cls, name: Optional[str] = None, config_path=None, config_mode=CONFIG_MODE_DEFAULT, **kwargs):
        """ Return an object  property  to be defined in a class 
        
        Exemple:
           
           ::
           
                def MyDevice(BaseDevice):
                   ref_temperature = StaticNode.prop(value=22.0
                
                MyDevice().ref_temperature.get()   
        """
        # config = cls.Config.parse_obj(kwargs)
        config = cls.parse_config(kwargs)
        return cls.Property(cls, cls.new, name, config=config, config_path=config_path, config_mode=config_mode)
    
    @classmethod
    def from_cfgfile(cls, 
            cfgfile: str, 
            key: Optional[str]= None, 
            path: Optional[Union[str,int]] = None, 
            prefix: str = '', 
            **kwargs
        ) -> '_BaseObject':
        """ Create the object from a configuration file 
        
        Args:
            cfgfile (str): Path to the config file, shall be relative to one of the $CFGPATH or absolute
                     path
            key (str, Optional): key of the device, if not given this is built from the path suffix and
                    the optional prefix
           
            path (None, str, False)"" the hierarchique path where to find the config data inside the file 
                    for instance 'a.b.c' will loock at cfg['a']['b']['c'] from the loaded config file 
                    If "" the config file define the device configuration from its root 
                    If None the first item of the config file is taken. 
                    Note that the path can be defined directly inside the cfgfile file name
                    in the form ``path/to/myconfig.yml[a.b.c]`` see :func:`load_config`

                                
            prefix (str, optional): add a prefix to the path name to build the device key. 
                    It is used only if key is None otherwhise ignored
        """
        
        allconf = load_config(cfgfile)
        allconf = path_walk_item(allconf, path)
        
        allconf.update(**kwargs)
               
        if key is None:
            if path and isinstance(path, str): 
                _, name = ksplit(path)           
                key = kjoin(prefix, name)
            else:
                # try with the path defined in file name 
                _, p = parse_file_name(cfgfile)
                if p:
                    key = p[-1]
                

        config = cls.Config.from_cfgfile( cfgfile, path )
        return cls(key = key, config=config, **kwargs)
            
    @property
    def key(self):
        """ key """
        return self._key
    
    @property
    def config(self):
        """  config """
        return self._config
    
    @property
    def localdata(self):
        """ localdata dictionary """
        return self._localdata
        
    @property
    def name(self):
        return ksplit(self._key)[1]    
    
    @classmethod
    def _builtin_objects(cls, SubCls=None):
        SubCls = SubCls or _BaseObject
        d = {}
        for sub in cls.__mro__:
            for k,v in sub.__dict__.items():
                if isinstance(v, (_BaseProperty)):
                    if issubclass(v._cls, SubCls):
                        d[k] = v._cls
        return d
    


    def _cash_all(self):
        """ cash all child _BaseProperty to real Objects """
        if not self.__all_cashed__:
            for sub in self.__class__.__mro__:
                for k,v in sub.__dict__.items():
                    if isinstance(v, (_BaseProperty)):
                        getattr(self, k)
            self.__all_cashed__ = True
    
    def _clear_all(self):
        """ clear all cashed intermediate objects """
        for k,v in list(self.__dict__.items()):
            if isinstance(v, (_BaseObject)):
                self.__dict__.pop(k)
        self.__all_cashed__ = False


class ChildrenCapability:
    _all_cashed = False
    _auto_build_object = AUTO_BUILD_DEFAULT

    def find(self, cls: Type[_BaseObject], depth: int = 0) -> Iterator:
        """ iterator on  children matching the given  class 
        
        ...note::
            
            The side effect of find is that all children will be built 

        Exemple::
            
            from pydevmgr_core import BaseNode
            list(   device.find( BaseNode, -1 ) ) # will return all nodes found in device and sub-devices, interface,
            etc...
        
        """
        if not self._all_cashed:
            self.build_all()
            self._all_cashed = True
        
        
        for obj in self.__dict__.values():
            if isinstance(obj, cls):
                yield obj
            if depth!=0 and isinstance(obj, ChildrenCapability):
                for sub in obj.find(cls, depth-1):
                    yield sub
            
    
            
    def build_all(self, depth:int =0) -> None:
        """ Build all possible children objects 
        
        Every single children will be built inside the object and will be cashed inside the obj.__dict__
        
        """
        for sub in self.__class__.__mro__:
                for k,v in sub.__dict__.items():
                    if isinstance(v, (_BaseProperty, _BaseObjDictProperty)):
                        obj = getattr(self, k)
                        if depth!=0 and isinstance(obj, ChildrenCapability):
                            obj.build_all(depth-1)

     
        if self._auto_build_object:
            for k,c in self.config:
                if isinstance( c, BaseConfig ):
                    obj = getattr( self, k)
                    if depth!=0 and isinstance(obj, ChildrenCapability):
                            obj.build_all(depth-1)

    def clear_all(self) -> None:
        """ Remove all instances of children object they a re rebuilt on demand """
        for k,v in list(self.__dict__.items()):
            if isinstance( v, _BaseObject ):
                del self.__dict__[k]
        
             
    

    def __getattr__(self, attr):   
        try:
            return object.__getattribute__(self, attr)
        except AttributeError as e:
            if self._auto_build_object:
                return auto_build(self, attr)
            else:
                raise e

    def children(self, cls: Optional[_BaseObject] = _BaseObject) -> Iterator:
        """ iter on children attribute name 

        Args:
            cls: the class which shall match the object. By default it will be all pydevmgr objects
            (:class:`pydevmgr_core._BaseObject`)

        Example::
                
            >>> l = [getattr(manager, name) for name in  manager.children( BaseDevice )]
            # is equivalent to 
            >>> l = list (manager.find( BaseDevice, 0 ))

            
        """
        if not self._all_cashed:
            self.build_all()
            self._all_cashed = True
        for name, obj in self.__dict__.items():
            if isinstance(obj, cls):
                yield name  
    
    
    def create_data_class( self, children: Iterator[str],  Base = None ) -> Type[BaseData]:
        """ Create a new data class for the object according to a child name list 

        This is a quick and durty way to create a data class dynamicaly. To be used mostly 
        in a manager or device with dynamic children. 
        Manager, Device, Interface child will be build from their defined Data class 
        Nodes will be of ``Any`` type and filled with ``None`` as default. 
        
        """       
        data_obj = {}
        for name in children:
            obj = getattr(self, name)
            if hasattr( obj, "get"):
                data_obj[name] = (NodeVar[Any], None)
            else:
                data_obj[name] = (obj.Data, obj.Data())
        Base  = self.Data if Base is None else Base 
        return create_model( "Data_Of_"+self.key, __base__= Base, **data_obj  )




    
class _BaseObjDict(dict, ChildrenCapability):
    
    def find(self, cls, depth: int = 0):
        for obj in self.values():
            if isinstance(obj, cls):
                yield obj
                
            if depth!=0 and isinstance(obj, ChildrenCapability):
                for sub in obj.find(cls, depth-1):
                    yield sub
    
    def build_all(self, depth=1):
        if depth==0: return 
        for obj in self.values():
            if isinstance(obj, ChildrenCapability):
                obj.build_all(depth-1)

        

class _BaseObjDictProperty:
    pass





from pydantic import BaseModel, Extra


class BaseCom: 
    Config = BaseModel
    class Config(BaseModel):
        class Config:
            extra  = Extra.forbid
    
    """ Define a BaseCom object 
    
    The Base does not do anything but is a place holder for what a com object should have
    A Com object can be set into a Device -> Interface -> Node for instance.           
    """ 
    def __init__(self, config=None, **kwargs):
        self._config = self.parse_config(config, **kwargs)
    
    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            return cls.Config( **{**__config__, **kwargs} )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")
        
    @property
    def config(self):
        return self._config 
    
    def connect(self):
        pass
    
    def disconnect(self):
        pass
    


from enum import Enum
from typing import Union, List, Optional, Type, Callable
class KINDS(str, Enum):
    PARSER = "Parser"
    NODE = "Node"
    RPC = "Rpc"
    DEVICE = "Device"
    INTERFACE = "Interface"
    MANAGER = "Manager"


class Parsers:
    pass
class Nodes:
    pass
class Rpcs:
    pass
class Devices:
    pass
class Interfaces:
    pass
class Managers:
    pass


object_loockup = {}
def get_class(kind: KINDS, type_: str, default=None) -> Type:
    try:
        return object_loockup[(kind, type_)]
    except KeyError:
        if default is None:
            raise ValueError(f"Unknown {kind!r} of type {type_!r}")
        else:
            return default
def get_parser_class(type_: str)-> Type:
    return get_class(KINDS.PARSER, type_)

def get_node_class(type_: str)-> Type:
    return get_class(KINDS.NODE, type_)

def get_rpc_class(type_: str)-> Type:
    return get_class(KINDS.RPC, type_)
        
def get_interface_class(type_: str)-> Type:
    return get_class(KINDS.INTERFACE, type_)

def get_device_class(type_: str)-> Type:
    return get_class(KINDS.DEVICE, type_)

def get_manager_class(type_: str)-> Type:
    return get_class(KINDS.DEVICE, type_)

def record_class(
         _cls_: Type =None, *, 
         overwrite: Optional[bool] = False, 
         type: Optional[str] =None, 
         kind: Optional[Union[KINDS,str]] =None
     ) -> Callable:
    """ record a new class by its kind and type 
    
    This can be used as decorator or function 
    """    
    if _cls_ is None:
        def obj_decorator(cls) -> Type:
            return record_class(cls, overwrite=overwrite, type=type)
        return obj_decorator
    else:
        cls = _cls_
        
    try:
        C = cls.Config
    except AttributeError:
        raise ValueError("Recorded class must have a Config class defined as attribute")
    
    if kind is None:        
        try:
            kind_field = C.__fields__['kind']    
        except (KeyError, AttributeError):
            raise ValueError("Config is missing 'kind' attribute or is not a BaseModel")
        else:
            kind = kind_field.default  
    
             
    if type is None:
        try:
            type_field = C.__fields__['type']    
        except (KeyError, AttributeError):
            raise ValueError("Config is missing 'type' attribute")
        else:
            type = type_field.default
            
    _record_class_as(kind_field.default, type, cls, overwrite=overwrite)
    return cls

def _record_class_as(kind, type, cls, overwrite=False):
    if not hasattr(cls, "Config"):
        raise ValueError("recorded class must have a Config attribute")
    if not overwrite and (kind, type) in object_loockup:
        raise ValueError(f"{kind} {type} object is already recorded, use overwrite keyword to replace")
    object_loockup[(kind, type)] = cls
    
    if kind == KINDS.PARSER:
        setattr(Parsers, type, cls)
    elif kind == KINDS.NODE:
        setattr(Nodes, type, cls)
    elif kind == KINDS.RPC:
        setattr(Rpcs, type, cls)
    elif kind == KINDS.INTERFACE:
        setattr(Interfaces, type, cls)    
    elif kind == KINDS.DEVICE:
        setattr(Devices, type, cls)
    elif kind == KINDS.MANAGER:
        setattr(Managers, type, cls)
    

def list_class(kind: Optional[KINDS] = None)-> Union[List[tuple],List[str]]:
    """ list all class names recorded class accessible with :func:`get_class` 
    
    Args:
        kind (str, KINDS, optional) :  If None the returned list is a list of (kind,type) tuple 
                                      Otherwise it musb a valid KINDS and the return list is a list 
                                      of type name 
    Return:
        classe_names (list):  list of tuple if kind=None or list of str if kind is given 
    """
    if kind is None:
        return list(object_loockup)
    else:
        return [t for k,t in object_loockup if k==kind]
            



""" a Yaml loader for pydevmgr used to provide some custom functionalities :

!math constructor to evaluate some mathematical expression 
!include to include a nested yaml file 

"""
import yaml 
from yaml import ScalarNode, MappingNode, SequenceNode
from py_expression_eval import Parser 
import os


CFGPATH = 'CFGPATH'  # name of the environment variable listing config dictionaries 

INCLUDE_KEY = 'include_files'
INCLUDE_TAG = '!include'
MATH_TAG = '!math'

math_parser  = Parser()
del Parser


def find_config(file_name):
    """ find a config file and return its absolute path 
    
    Args:
        file_name (str): config file name. Should be inside one of the path defined by the $CFGPATH 
        environment variable. Alternatively the file_name can be an abosolute path       
    """
    path_list = os.environ.get(CFGPATH , '.').split(':')
    for directory in path_list[::-1]:
        path = os.path.join(directory, file_name)
        if os.path.exists(path):
            return  path
    raise ValueError('coud not find config file %r in any of %s'%(file_name, path_list))


class PydevmgrLoader(yaml.CLoader):
    """ yaml loader with the !include constructor 
    
    !include constructor can be built a scalar string (file path) or a mapping 
    
    - if a file path : it shall be a a yaml configuration file, it is loaded and included in place 
    - if a mapping : the files to include shall be in a includes list of files path 
    the contents of files (which shall be a mapping) are included inside the mapping a value  
    inside the last file will erase the previous one. Ultimatly the values can be overwriten 
    by in the current mapping 
    
    Exemple
    -------

    ---
    name : test
    config: !include path/to/config.yml
    

    --- 
    name: test 
    config: !include
     include_files: [path/to/config1.yml, path/to/config2.yml]
     overwriten_key: 34.5 
    
    """
    find_config = staticmethod(find_config)
   	


def include_constructor(loader, node):

    if isinstance(node, ScalarNode):

        with open(find_config(loader.construct_scalar(node)), 'r') as f:
            return yaml.load(f, PydevmgrLoader)

    
    elif  isinstance(node, MappingNode):
        data = loader.construct_mapping(node)
        try:
            data.pop(INCLUDE_KEY)
        except KeyError:
            return data 
    
        files = _find_includes(loader, node) 
        for file in files[::-1]:
            with open(find_config(file), 'r') as f:
                src = yaml.load(f, PydevmgrLoader)
                if not isinstance(src, dict):
                    raise ValueError("included file inside a mapping is expected to be a mapping")
                _merge_dictionary(data, src)		

        return data		

    else:
        raise ValueError("Bad argument for !include flag constructor")	



def math_constructor(loader, node):
    return math_parser.parse(loader.construct_scalar(node)).evaluate({})

    # filename = os.path.join(self._root, self.construct_scalar(node))

    # with open(filename, 'r') as f:
    #     return yaml.load(f, Loader)    
    # return     

PydevmgrLoader.add_constructor(INCLUDE_TAG, include_constructor)
PydevmgrLoader.add_constructor(MATH_TAG, math_constructor)


def _find_includes(loader, node):

    for k,v in node.value:
        if isinstance(k, ScalarNode) and k.value == INCLUDE_KEY:
            if isinstance(v, ScalarNode):				
                return [loader.construct_scalar(v)]
            if isinstance(v, SequenceNode):
                return loader.construct_sequence(v) 
            else:
                raise ValueError(f"Bad argument for {INCLUDE_KEY!r} constructor")
    return []	


def _merge_dictionary(dst, src):
    stack = [(dst, src)]
    while stack:
        current_dst, current_src = stack.pop()
        for key in current_src:
            if key not in current_dst:
                current_dst[key] = current_src[key]
            else:
                if isinstance(current_src[key], dict) and isinstance(current_dst[key], dict) :
                    stack.append((current_dst[key], current_src[key]))
                #else:
                #    current_dst[key] = current_src[key]
    




#Loader = yaml.SafeLoader




if __name__ == "__main__":

    includes = [
("test_include1.yml", 
"""---
ctrl:
    min: 0
    max: 10
    velocity: 3

"""),

("test_include2.yml",
"""---
ctrl:
    velocity: 5.0
"""), 


("test_include3.yml",
"""ctrl: !include test_include4.yml
"""),

("test_include4.yml",
"""---
min: -180
max:  180
velocity: 1
"""),

]



    root = """---
motor1: !include 
    include_files: [test_include1.yml, test_include2.yml]
    name: This is a test of the include fonctionality 
    ctrl:
        max: 5

motor2: !include test_include3.yml
x: !math 3+5
y: !math sin(PI/2.0)
"""
    #print( MyModel.parse_obj(yaml.load(t)) )



    tmp_dir = os.path.join('/', 'tmp')
    os.environ[CFGPATH] = tmp_dir
    for file, txt in includes:
        with open( os.path.join(tmp_dir , file), 'w') as g:
            g.write(txt)

    
    obj1 = yaml.load(root, Loader=PydevmgrLoader)
    assert obj1['motor1']['ctrl']['min'] == 0 
    assert obj1['motor1']['ctrl']['max'] == 5
    assert obj1['motor1']['ctrl']['velocity'] == 5.0
    assert obj1['motor2']['ctrl']['velocity'] == 1.0
    assert obj1['x'] == 8
    assert obj1['y'] == 1.0

    obj2 = yaml.load("!include test_include1.yml", Loader=PydevmgrLoader)
    assert obj2['ctrl']['min'] == 0 
    print("OK")







import weakref
from ._core_base import (_BaseObject, _BaseProperty,
                          BaseData, 
                         ChildrenCapabilityConfig,  ChildrenCapability
                        )
                         
from ._class_recorder import get_interface_class, get_class, record_class, KINDS

from ._core_com import BaseCom
from ._core_node import BaseNode
from ._core_rpc import BaseRpc
from ._core_obj_dict import ObjDict

from .. import io
from typing import Optional, Iterable, Union, List, Dict, Callable
#  ___ _   _ _____ _____ ____  _____ _    ____ _____ 
# |_ _| \ | |_   _| ____|  _ \|  ___/ \  / ___| ____|
#  | ||  \| | | | |  _| | |_) | |_ / _ \| |   |  _|  
#  | || |\  | | | | |___|  _ <|  _/ ___ \ |___| |___ 
# |___|_| \_| |_| |_____|_| \_\_|/_/   \_\____|_____|
# 


class BaseInterfaceConfig(_BaseObject.Config, ChildrenCapabilityConfig):
    """ Config for a Interface """
    kind: KINDS = KINDS.INTERFACE.value
    type: str = "Base"     
    
        
class InterfaceProperty(_BaseProperty):    
    fbuild = None
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self

    def __call__(self, func):
        """ The call is used has fget decorator """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, interface):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, interface)            

class InterfaceDict(ObjDict):
    class Config(ObjDict.Config):
        kinds = set([KINDS.INTERFACE])
        default_kind= KINDS.INTERFACE 



@record_class # we can record this type because it should work as standalone        
class BaseInterface(_BaseObject, ChildrenCapability):
    """ BaseInterface is holding a key, and is in charge of building nodes """    
    
    _subclasses_loockup = {} # for the recorder 
    
    Config = BaseInterfaceConfig
    Property = InterfaceProperty
    Data = BaseData
    Node = BaseNode
    Rpc = BaseRpc   
    Dict = InterfaceDict
    
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] = None,            
           **kwargs
        ) -> None:        
        
        super().__init__(key, config=config, **kwargs)  
        if self._localdata is None:
            self._localdata = {}
    

    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()        

                    

                

    



from pydevmgr_core._core_objects._class_recorder import get_class, KINDS
from pydantic import BaseModel, ValidationError
from pydantic.fields import ModelField
from typing import Optional, TypeVar, Generic 
#from .. import io 

from ..io import load_config

RecVar= TypeVar('RecVar')
GenVar = TypeVar('GenVar')
GenConfVar = TypeVar('GenConfVar')


def _isdefault(field):
    try:
        return issubclass(field.type_, Defaults)
    except Exception:
        return False
                

def _default_walk_unset(  default: BaseModel, new: BaseModel ):
    
    if not isinstance( new, BaseModel ):
        return 
    
    fields = default.__fields__
    
    for k, v in default:
        if k in fields and _isdefault(fields[k]):
            sub = getattr(new, k)
            _default_walk_unset(v, sub)
        else:
            if not k in new.__fields_set__:
                # add the value directly inside __dict__ 
                # This is necessary otherwhise it will not work recursively 
                # Can be a huge draw back but not sure how to fix it, adding in __fields_set__ break the recursibility
                new.__dict__[k] = v
                # new.__fields_set__.add(k)


def _default_walk_set(  default: BaseModel, new: BaseModel ):
    
    if not isinstance( new, BaseModel ):
        return 
    
    
    # if new.cfgfile:
    #     cfgfile_data = io.load_config(new.cfgfile) 
    #     cfgfile_model = new.__class__.parse_obj(cfgfile_data)
    #     _default_walk_set( default, cfgfile_model )
    #     default =  cfgfile_model 


    fields = default.__fields__
    
    for k, v in default:
        if k in fields and _isdefault(fields[k]):
            sub = getattr(new, k)
            _default_walk_set(v, sub)
        else:
            if not k in new.__fields_set__ or ("__default__"+k) in new.__fields_set__:
                # add the value directly inside __dict__
                new.__dict__[k] = v
                # If the field has been set by the walker the __key shall be also in __fields__set__
                # so we need to update it to keep it recursive. The draw back is keys liek "__key" in __fields__set__
                # but should not be a problem as it is used mostly for excluding stuff 
                new.__fields_set__.add("__default__"+k)
                new.__fields_set__.add(k)


class Defaults(Generic[RecVar]):
    """ Make the value of a default submodel the default values of the incoming payload 
    """
    _walker = _default_walk_set
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        #field_schema.update()
        pass

    @classmethod
    def validate(cls, v, field: ModelField):
        if not field.sub_fields:
            # Generic parameters were not provided so we don't try to validate
            # them and just return the value as is
            return v
        if len(field.sub_fields)!=1:
            raise ValidationError(['to many field Defaults require and accept only one argument'], cls)
            
        
        val_f = field.sub_fields[0]
        errors = []
        
        valid_value, error = val_f.validate(v, {}, loc='value')
        

        if error:
            errors.append(error)
        if errors:
            raise ValidationError(errors, cls)
        if field.default is not None:
            cls._walker(field.default, valid_value) 
        # Validation passed without errors, return validated value
        return valid_value
    
    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'


class Defaults2(Defaults):
    _walker = _default_walk_unset



class GenObject(Generic[GenVar]):
    """ Make the value of a default submodel the default values of the incoming payload 
    """
    _child_kind = None    
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        #field_schema.update()
        pass
    
    @classmethod
    def validate_child(cls, kind, type_):
        if kind is not None and kind!=cls._child_kind:
            raise ValueError(f"expecting a {cls._child_kind} configuration got a {kind}")
        if type_ is None:
            raise ValueError(f"{cls.__name__} must define a type")
        return cls._child_kind, type_
    
    @classmethod
    def validate(cls, v, field: ModelField):
                    
        if field.sub_fields:
            if len(field.sub_fields)!=1:
                raise ValidationError(['to many field GenDevice require and accept only one argument'], cls)
        

            val_f = field.sub_fields[0]
            errors = []
        
            valid_value, error = val_f.validate(v, {}, loc='value')
        
        

            if error:
                errors.append(error)
            if errors:
                raise ValidationError(errors, cls)
        else:
            valid_value = v

        if isinstance(valid_value, BaseModel):
            kind = getattr(valid_value, 'kind', None)
            cls.validate_child(kind, getattr(valid_value, 'type', None))

        elif isinstance(valid_value, dict):

            kind, type_ = cls.validate_child( 
                   valid_value.get('kind', None), 
                   valid_value.get('type', None)
            )
            
                
            Obj = get_class(kind, type_)
            valid_value = Obj.Config.parse_obj(valid_value)
        else:
            raise ValueError(f"Unexpected value for {cls.__name__}")
        # Validation passed without errors, return validated value
        return valid_value
    
    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'


class GenManager(GenObject):   
    _child_kind = KINDS.MANAGER

class GenDevice(GenObject):
    _child_kind = KINDS.DEVICE

class GenInterface(GenObject):
    _child_kind = KINDS.INTERFACE

class GenNode(GenObject):
    _child_kind = KINDS.NODE

class GenRpc(GenObject):
    _child_kind = KINDS.RPC

class GenParser(GenObject):
    _child_kind = KINDS.PARSER




class _BaseModelGenConf(BaseModel):
    class Config:
        extra = "allow"

class GenConf(Generic[GenConfVar]):
    """ Parse a Model. If the input is a string it is interpreted as a path to a config file 

    The model is loaded from the config file which must have an absolute path or a path relative 
    to one of the path defined in the $CFGPATH env varaible 
    
    A Node input will be interpreted as an empty dictionary (Model built without arguments). It will 
    fail if model has required entries 
    Example: 

    ::
        
        form pydevmgr_core import BaseDevice, GenConf

        class Config(BaseDevice.Config):
            child :  GenConf[BaseDevice.Config] = BaseDevice.Config()

        c = Config( child={})
        c = Config( child="/path/to/my/file.yml")
        

    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        pass
    
    
    @classmethod
    def validate(cls, value, field: ModelField):
                    
        if field.sub_fields:
            if len(field.sub_fields)!=1:
                raise ValidationError(['to many field GenConf require and accept only one argument'], cls)
                
        

            Model = field.sub_fields[0]
            if isinstance( Model, ModelField):
                Model = Model.type_
            
            if not hasattr(Model, "parse_obj"):
                 raise ValueError(f"field of GenConf is not a BaseModel but a {type(Model)}")
        else:
            Model = _BaseModelGenConf
            
        if isinstance( value, str):
            value = load_config( value )
        # if value is None:
        #     return Model()
        return Model.parse_obj(value)

    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'



if __name__ == "__main__":
    from pydantic import BaseModel
    from pydevmgr_core import BaseDevice, record_class
    from typing import Dict 
    
    @record_class
    class Toto(BaseDevice):
        class Config(BaseDevice.Config):
            type = "Toto"
            voltage: float = 12.0
    
    class C(BaseModel):
        n1: int = 0
        n2: int = 0

    class B(BaseModel):
        x: float = 0.0
        y: float = 0.0
        c1: Defaults[C] = C(n1=1)
        c2: Defaults[C] = C(n1=2)
        c3: C = C(n1=3)

    RB = Defaults[B]
    class A(BaseModel):
        b1: Defaults[B] = B(y=9)
        b2: Defaults[B] = B(y=8, c1=C(n1=100))
        
        devices: Dict[str,GenDevice] = {}

    a =A(b1={'x':1.0, 'c1':{'n2':10},  'c2':{'n2':20}, 'c3':{'n2':30}},    b2={'c1':{}}, 
            devices={ 'toto':{'type':'Toto'}, 'toto2':Toto.Config()}
    
        )

    print(a)
    
    assert a.b1.x == 1.0
    assert a.b1.y == 9.0 
    assert a.b1.c1.n2 == 10
    assert a.b1.c1.n1 == 1
    assert a.b1.c2.n1 == 2
    assert a.b1.c3.n1 == 0 # c3 is not a Defaults 

    assert a.b2.c1.n1 == 100
    
    assert "y" in a.b1.dict(exclude_unset=True)
  



from enum import Enum 

class _S(str):
    def __init__(self, s, style=""):
        super().__init__(s)
        self.style = style
        
class STYLE(str, Enum):
    """ A collection of style IDs derived from GROUPs in pydevmgr + extra stuff """
    IDL       = "IDL"
    WARNING   = "WARNING"
    ERROR     = "ERROR"
    OK        = "OK"
    NOK       = "NOK"
    BUZY      = "BUZY"
    UNKNOWN   = "UNKNOWN"
    NORMAL    = "NORMAL"
    ODD       = "ODD"
    EVEN      = "EVEN"
    ERROR_TXT = "ERROR_TXT"
    OK_TXT    = "OK_TXT"
    DIFFERENT = "DIFFERENT"
    SIMILAR   = "SMILAR"

    
""" Associate STYLE IDs to qt stylSheet """
_qt_style_loockup = {
    STYLE.NORMAL  : "background-color: white;",
    STYLE.IDL     : "background-color: white;",
    STYLE.WARNING : "background-color: #ff9966;",
    STYLE.ERROR   : "background-color: #cc3300;",
    STYLE.OK      : "background-color: #99cc33;",
    STYLE.NOK     : "background-color: #ff9966;",
    STYLE.BUZY    : "background-color: #ffcc00;",
    STYLE.UNKNOWN : "",
    STYLE.ODD     : "background-color: #E0E0E0;",
    STYLE.EVEN    : "background-color: #F8F8F8;",
    STYLE.ERROR_TXT : "color: #cc3300;",
    STYLE.OK_TXT : "color: black;",
    STYLE.DIFFERENT : "color: #cc3300;",
    STYLE.SIMILAR: "color: black;",
}
class STYLE_DEF:
    pass

for _N, _S in _qt_style_loockup.items():
    setattr(STYLE_DEF, _N, _S)
del _N, _S, _qt_style_loockup
    
def get_style(style):
    """ return the style of a given style name """
    return getattr(STYLE_DEF, style, "")



from ._core_objects import NodesReader,  DictReadCollector, BaseNode
import time


class Waiter:   
    """ Object use to wait for some condition to be True 
    
    The only method of this object is :meth:`Waiter.wait`
    
    Args:
        node_or_func_list (iterable): an iterable of nodes or callable. Or a single node 
                    waiter.wait will wait until all node.get and callable return True (by default)
        logic (string, optional):
              "all_true"  : stop waiting when all check function return True (default)
              "all_false" : stop waiting when all check function return False
              "any_true"  : stop waiting when one of the check function return True
              "any_false" : stop waiting when one of the check function return False
            
        period : float, optional default is 0.1 second
            the sleep time in second in between checks
        timeout: float, optional second, default is 60
            if time since the call of the function exceed timeout an
            ValueError is raised.
            timeout is in second as well
        stopper (callable, optional): this function is called at each cycle, it should raise a StopIteration
                                      or any error. 
        data (None, dict, optional):  If given input nodes values are taken from the 
               data dictionary which is expected to be updated in someother place.
        
    Example:
        
    ::
        
        >>> wait_moving = Waiter([motor.stat.is_moving])
        >>> wait_moving.wait()
        
    wait return True and can be used as a trigger for a :class:`Download` object :
    
    :: 
    
        >>> data = {} 
        >>> pos_update = Downloader(data, [motor.stat.pos_actual], triger=wait_moving.wait)
        >>> t = Thread(target=pos_update.runner(0.5))
        >>> t.start()
        
    In the exemple above the thread will update the data dictionary 
    every 0.5 seconds with motor position only when the motor is moving. 
            
    Attributes:
        period (float): conditions are checks every period seconds
        timeout (float): timeout in second, a RuntimeError is raised if conditions
                         are still false after timeout.
    """
    def __init__(self, node_or_func_list, logic="all_true", period=0.1, timeout=60, stopper=lambda:None, data=None, ):        
        nodes = []
        functions = []
        if not hasattr(node_or_func_list, "__iter__"): 
            node_or_func_list = [node_or_func_list]
        
        for f in node_or_func_list:
            if f is None: continue
            if isinstance(f, BaseNode):
                nodes.append(f)
            else:
                functions.append(f)
        
        self._functions = functions
            
        
        try:
            self.check_func, self.check_nodes = _logic_loockup[logic]
        except KeyError:
            raise ValueError("undefined logic %r must be one of %s"%(logic, ",".join(_logic_loockup.keys())))
        
        if data is None:
            reader = NodesReader(nodes)
        else:
            reader = DictReadCollector(data)
            for n in nodes:
                reader.add(n)           
        self.nodes = nodes               
        self.reader = reader    
        self.period  = period
        self.timeout = timeout
        self.stopper = stopper
        
    def wait(self, lag=0.0):
        """ run the wait condition 
        
        Args:
            lag (float, optional): Add a flat time lag (in second) before starting to wait. 
                        This could be used to make sure that the last operation has been digested by server. 
                             
                        Bellow the lag is used to make sure that when ``wait`` starts the motor is moving
                              
                        ::
                                
                            >>> waiter = Waiter([mgr.motor1.stat.is_standstill])
                            >>> mgr.motor1.move_rel(1.0, 0.25)
                            >>> waiter.wait( 0.1 )                
        """
        check_nodes     = self.check_nodes
        check_func     = self.check_func
        reader  = self.reader
        
        functions = self._functions         
        timeout   = self.timeout
        period    = self.period
        stopper   = self.stopper 
        s_time = time.time()
        if lag>0.0:
            time.sleep(lag)
        
        data_nodes = {}
        def check():
            # do the function first if it pass download the node values and check values 
            q = check_func(functions)
            if q:
                reader.read(data_nodes)
                return check_nodes( [data_nodes[n] for n in self.nodes] )
            return False
        
        # do the function first 
        while not check():
            stopper() # raise a SopIteration if needed to stop. The StopIteration shall be catched 
                      # at higher level like in Measure.run or Download.run  
                                
            if (time.time()-s_time)>timeout:
                raise RuntimeError('wait timeout')
            time.sleep(period)
        
        return True
    

def wait(node_or_func_list, logic="all_true", period=0.1, timeout=60, lag=0.0, stopper=lambda:None, data=None):
    """ wait until a list of function return True

    Args:
        node_or_func_list (iterable): an iterable of nodes of callable.  Or a single node
                    wait will wait until all node.get and callable return True (by default)
                    
        logic (str, optional): 
              "all_true"  : stop waiting when all check function return True (default)
              "all_false" : stop waiting when all check function return False
              "any_true"  : stop waiting when one of the check function return True
              "any_false" : stop waiting when one of the check function return False
            
        period (float, optional): default is 0.1 second
            the sleep time in second in between checks
        timeout (float, optional): timeout in second, a RuntimeError is raised if conditions
                         are still false after timeout
        lag (float, optional): Add a flat time lag (in second) before checking nodes. 
                    This could be used to make sure that the last operation has been digested by server. 
                         
                    Bellow the lag is used to make sure that when ``wait`` starts the motor is moving
                          
                    ::
                             
                        >>> mgr.motor1.move_rel(1.0, 0.25)
                        >>> wait( [mgr.motor1.stat.is_standstill], lag=0.1 )
                            
        stopper (callable, optional): this function is called at each cycle, it should raise a StopIteration
                                      or any error. 
        data (None, dict, optional):  If given input nodes values are taken from the 
               data dictionary which is expected to be updated in someother place.

    Example:
        Wait until a motor initialised and an other custom function
        
        ::

            > def camera_ready():
            >   # <- do stuff ->
            >   return True # if camera is ready

            > wait( [motor.is_initialised, camera_ready] )

        Or something like
        
        ::
        
            > is_arrived = lambda : abs(motor.stat.pos_actual.get()-3.4)<0.01
            > wait( [is_arrived, camera_ready])

    """
    Waiter(node_or_func_list, logic=logic, period=period, timeout=timeout, stopper=stopper, data=data).wait(lag)

def _all_true(functions):
    """ all_true(lst) -> return True if all function  list return True """
    for func in functions:
        if not func():
            return False
    return True

def _all_false(functions):
    """ all_false(lst) -> return True if all function in the list return False """

    for func in functions:
        if func():
            return False
    return True
    
def _any_true(functions):
    """ any_true(lst) -> return True if one of the function in the list return True """
    for func in functions:
        if func():
            return True        
    return False

def _any_false(functions):
    """ any_false(lst) -> return True if one of the function in the list return False """
    for func in functions:
        if not func():
            return True
    return False


""" Used in wait to define the logic applied """
_logic_loockup = {
    "all_true"  : (_all_true,  all),
    "all_false" : (_all_false, lambda l: not any(l) ), 
    "any_true"  : (_any_true,  any),
    "any_false" : (_any_false,  lambda l: not all(l))
}




from ._class_recorder import (get_class, get_rpc_class, get_node_class, get_interface_class, 
                              get_device_class, get_manager_class, record_class, KINDS, 
                              list_class, Nodes, Rpcs, Devices, Managers, Parsers, Interfaces 
                              )
from ._core_base import (kjoin, ksplit, reconfig, build_yaml, load_and_build, BaseData,
                        open_object, _BaseObject, path_walk_item , path_walk_attr, path)    
from ._core_node import (BaseNode,
                         nodeproperty,  node, 
                          NodesReader, NodesWriter, 
                          DictReadCollector, DictWriteCollector, 
                          BaseReadCollector, BaseWriteCollector, 
                          to_node_class, 
                          new_node                      
                         )
from ._core_node_alias import (NodeAlias, NodeAlias1, nodealiasproperty, nodealias, nodealias1, to_nodealias_class)
from ._extra_nodes import StaticNode, LocalNode

from ._core_rpc import RpcError, BaseRpc, RpcProperty, rpcproperty, to_rpc_class
from ._core_interface import BaseInterface, InterfaceProperty  

from ._core_device import BaseDevice, DeviceProperty, open_device
from ._core_manager import BaseManager, ManagerProperty, open_manager
from ._core_com import BaseCom
from ._core_model_var import NodeVar, NodeVar_R, NodeVar_W, NodeVar_RW, StaticVar

from ._core_pydantic import Defaults, GenDevice, GenManager, GenInterface, GenNode, GenConf


from ._core_parser import *
